import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractConstrainedMinimisation


class HAGER4(AbstractConstrainedMinimisation):
    """
    A nonlinear optimal control problem, by W. Hager.

    NOTE: The solution for x given in the article below by Hager has
    a typo. On the interval [1/2, 1], x(t) = (exp(2t) + exp(t))/d. In
    other words, the minus sign in the article should be a plus sign.

    Source: problem P4 in
    W.W. Hager,
    "Multiplier Methods for Nonlinear Optimal Control",
    SIAM J. on Numerical Analysis 27(4): 1061-1080, 1990.

    SIF input: Ph. Toint, April 1991.

    classification OLR2-AN-V-V

    Default N = 2500 from SIF file
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Problem parameters - using the default value from SIF
    n_param: int = 2500  # Number of discretized points in [0,1]

    # Total number of variables: x(0) to x(N) plus u(1) to u(N)
    n: int = 2 * n_param + 1  # 5001

    # Number of constraints: N constraints S(i) (x(0) is fixed, not a constraint)
    m: int = n_param  # 2500

    def objective(self, y: Array, args) -> Array:
        """Compute the objective function."""
        n = self.n_param
        h = 1.0 / n

        # Extract variables
        x = y[: n + 1]  # x(0) to x(N)
        u = y[n + 1 :]  # u(1) to u(N)

        # Compute time-dependent quantities
        t = jnp.arange(n + 1, dtype=jnp.float64) * h
        z = jnp.exp(-2.0 * t)

        # Constants for element calculations
        a = -0.5 * z
        b = a * (t + 0.5)
        c = a * (t * t + t + 0.5)

        # Differences for scaling
        da = a[1] - a[0]
        db = b[1] - b[0]
        dc = c[1] - c[0]

        # From SIF: 1/H = N, so SCDB = DB * (1/H) = DB * N
        # and SCDC = DC * (1/2HSQ) = DC * 0.5 * N^2
        scda = 0.5 * da
        scdb = db * n  # DB * (1/H) where 1/H = N
        scdc = dc * (0.5 * n * n)  # DC * (1/2HSQ) where 1/2HSQ = 0.5*N^2

        # ELT elements (vectorized)
        # Element parameters - use z[i-1] for element i
        # For elements 1 to N, we need z[0] to z[n-1]
        d = scda * z[:-1]  # z[0] to z[n-1]
        e = scdb * z[:-1]
        f = scdc * z[:-1]

        # ELT: D*X*X + E*X*(Y-X) + F*(Y-X)**2
        # where X = x(i), Y = x(i-1)
        x_curr = x[1:]  # x[1] to x[n] - this is X in the element
        x_prev = x[:-1]  # x[0] to x[n-1] - this is Y in the element
        diff = x_prev - x_curr  # Y - X

        elt = d * x_curr * x_curr + e * x_curr * diff + f * diff * diff
        obj = jnp.sum(elt)

        # u[i]^2 terms scaled by h/2
        obj += jnp.sum(u**2) * (h / 2.0)

        return obj

    def constraint(self, y: Array):
        """Compute the equality and inequality constraints."""
        n = self.n_param
        h = 1.0 / n

        # Extract variables
        x = y[: n + 1]  # x(0) to x(N)
        u = y[n + 1 :]  # u(1) to u(N)

        # Compute time-dependent quantities
        t = jnp.arange(n + 1, dtype=jnp.float64) * h

        # Equality constraints
        # S(i) constraints for i = 1 to N (vectorized)
        # S(i): (1/h - 1)*x(i) + (-1/h)*x(i-1) - exp(t(i))*u(i) = 0
        # Note: x(0) is fixed to XX0, not handled as a constraint
        eti = jnp.exp(t[1:])  # exp(t(1)) to exp(t(N))
        eq_constraints = (1.0 / h - 1.0) * x[1:] + (-1.0 / h) * x[:-1] - eti * u

        # No inequality constraints
        ineq_constraints = None

        return eq_constraints, ineq_constraints

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """x(0) is fixed to XX0 via bounds, u(i) bounded above by 1.0."""
        n = self.n_param
        lower = jnp.full(self.n, -jnp.inf)
        upper = jnp.full(self.n, jnp.inf)

        # Fix x(0) to XX0 using bounds
        e = jnp.exp(1.0)
        xx0 = (1.0 + 3.0 * e) / (2.0 - 2.0 * e)
        lower = lower.at[0].set(xx0)
        upper = upper.at[0].set(xx0)

        # u(i) <= 1.0 for i = 1 to N
        upper = upper.at[n + 1 :].set(1.0)

        return lower, upper

    @property
    def y0(self) -> Array:
        """Initial guess for the optimization problem."""
        y = jnp.zeros(self.n)
        # x(0) starts at specific value based on constants
        e = jnp.exp(1.0)
        xx0 = (1.0 + 3.0 * e) / (2.0 - 2.0 * e)
        y = y.at[0].set(jnp.array(xx0, dtype=y.dtype))
        return y

    @property
    def args(self):
        """Additional arguments for the objective and constraint functions."""
        return None

    @property
    def expected_result(self) -> Array:
        """Expected result of the optimization problem."""
        # Not explicitly given in the SIF file
        return jnp.zeros(self.n)

    @property
    def expected_objective_value(self) -> Array:
        """Expected value of the objective at the solution."""
        # From SIF file comments: SOLTN(1000) = 2.794244187
        return jnp.array(2.794244187)



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: Needs verification against another CUTEst interface
class HIMMELBG(AbstractUnconstrainedMinimisation):
    """The HIMMELBG function.

    A 2-variable problem by Himmelblau.

    Source: problem 33 in
    D.H. Himmelblau,
    "Applied nonlinear programming",
    McGraw-Hill, New-York, 1972.
    See Buckley#87 (p. 67)

    SIF input: Ph. Toint, Dec 1989.
    Classification: OUR2-AN-2-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y, args):
        del args
        x1, x2 = y
        exponential = jnp.exp(-x1 - x2)
        quadratic = 2 * x1**2 + 3 * x2**2
        return exponential * quadratic

    @property
    def y0(self):
        return jnp.array([0.5, 0.5])

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        return None

    @property
    def expected_objective_value(self):
        return jnp.array(0.0)



"""A penalty problem by Gill, Murray and Pitfield.

It has a dense Hessian matrix.

Source:  problem 114 (p. 81) in
A.R. Buckley,
"Test functions for unconstrained minimization",
TR 1989CS-3, Mathematics, statistics and computing centre,
Dalhousie University, Halifax (CDN), 1989.

SIF input: Nick Gould, Dec 1990.

classification OUR2-AY-V-0
"""

import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class PENALTY3(AbstractUnconstrainedMinimisation):
    """A penalty problem by Gill, Murray and Pitfield."""

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})
    n: int = 200  # Default to n=200 (N/2=100)

    @property
    def y0(self):
        """Initial guess: (1, -1, 1, -1, ..., 1, -1)."""
        x = jnp.ones(self.n)
        # Set every other element to -1
        x = x.at[1::2].set(-1.0)
        return x

    @property
    def args(self):
        """No additional arguments."""
        return None

    def objective(self, y, args):
        """Compute the objective function."""
        del args  # Not used

        x = y
        n = self.n
        n_half = n // 2
        rn = float(n)
        a = 0.001

        # Initialize objective value
        obj = 0.0

        # A group: constant -a
        obj += -a

        # REXP group: sum over i=1 to n-2
        # Vectorized: u1 = x[i] + 2.0 * x[i+1] + 10.0 * x[i+2]
        u1 = x[:-2] + 2.0 * x[1:-1] + 10.0 * x[2:]
        u3 = x[n - 1]
        u1m1 = u1 - 1.0
        rexp_sum = jnp.sum(a * u1m1 * u1m1 * jnp.exp(u3))
        obj += rexp_sum

        # SEXP group: sum over j=1 to n-2
        # Vectorized: u2 = 2.0 * x[j] + x[j+1]
        u2 = 2.0 * x[:-2] + x[1:-1]
        u3 = x[n - 2]
        u2m3 = u2 - 3.0
        sexp_sum = jnp.sum(a * u2m3 * u2m3 * jnp.exp(u3))
        obj += sexp_sum

        # RS group: double sum over i,j=1 to n-2
        # Vectorized using outer product
        u1 = x[:-2] + 2.0 * x[1:-1] + 10.0 * x[2:]
        u2 = 2.0 * x[:-2] + x[1:-1]
        u1m1 = u1 - 1.0
        u2m3 = u2 - 3.0
        # Outer product creates (n-2) x (n-2) matrix
        rs_sum = jnp.sum(a * jnp.outer(u1m1 * u1m1, u2m3 * u2m3))
        obj += rs_sum

        # T**2 group: double sum over i,j=1 to n
        # Vectorized using outer product
        x2mrn = x * x - rn
        t_sum = jnp.sum(jnp.outer(x2mrn, x2mrn))
        obj += t_sum

        # U group: sum over i=1 to n/2
        xim1 = x[:n_half] - 1.0
        u_sum = jnp.sum(xim1 * xim1)
        obj += u_sum

        return obj

    @property
    def expected_result(self):
        """Expected optimal solution."""
        # The expected objective value is approximately 0.001
        # No specific expected result provided in SIF
        return None

    @property
    def expected_objective_value(self):
        """Expected objective value is approximately 0.001."""
        return jnp.array(0.001)



import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractNonlinearEquations


class OSCIGRNE(AbstractNonlinearEquations):
    """
    The roots of the gradient of Yurii Nesterov's "oscillating path" problem
    Nonlinear equations version

    SIF input: Nick Gould, June 2011.

    classification NOR2-AN-V-V
    """

    n: int = 100000
    rho: float = 500.0  # Florian Jarre's value
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def __init__(self, n: int = 100000, rho: float = 500.0):
        self.n = n
        self.rho = rho

    def num_residuals(self) -> int:
        """Number of residuals equals number of variables."""
        return self.n

    def starting_point(self) -> Array:
        """Return the starting point for the problem."""
        x0 = jnp.ones(self.n, dtype=jnp.float64)
        x0 = x0.at[0].set(-2.0)
        return x0

    def residual(self, x: Array, args) -> Array:
        """Compute the residual vector."""
        residuals = jnp.zeros(self.n, dtype=jnp.float64)

        # G(1) = 0.5 * X(1) - 0.5 + B1
        # B1 element: P * (V - 2.0 * U^2 + 1.0) * U
        # with V = X(2), U = X(1), P = -4*RHO
        b1_p = -4.0 * self.rho
        b1_val = b1_p * (x[1] - 2.0 * x[0] ** 2 + 1.0) * x[0]
        residuals = residuals.at[0].set(0.5 * x[0] - 0.5 + b1_val)

        # For i = 2 to N-1
        # G(i) = A(i) + B(i)
        # A(i): P * (V - 2.0 * U^2 + 1.0) with V = X(i), U = X(i-1), P = 2*RHO
        # B(i): P * (V - 2.0 * U^2 + 1.0) * U with V = X(i+1), U = X(i), P = -4*RHO
        if self.n > 2:
            a_p = 2.0 * self.rho
            b_p = -4.0 * self.rho

            # Vectorized computation for middle elements
            i_mid = jnp.arange(1, self.n - 1)
            a_vals = a_p * (x[i_mid] - 2.0 * x[i_mid - 1] ** 2 + 1.0)
            b_vals = b_p * (x[i_mid + 1] - 2.0 * x[i_mid] ** 2 + 1.0) * x[i_mid]
            residuals = residuals.at[i_mid].set(a_vals + b_vals)

        # G(N) = A(N)
        # A(N): P * (V - 2.0 * U^2 + 1.0) with V = X(N), U = X(N-1), P = 2*RHO
        if self.n > 1:
            a_n_p = 2.0 * self.rho
            a_n_val = a_n_p * (x[-1] - 2.0 * x[-2] ** 2 + 1.0)
            residuals = residuals.at[-1].set(a_n_val)

        return residuals

    @property
    def y0(self) -> Array:
        """Initial guess for the optimization problem."""
        return self.starting_point()

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> Array:
        """Expected result of the optimization problem."""
        # Solution should satisfy F(x*) = 0
        return jnp.zeros(self.n, dtype=jnp.float64)

    @property
    def expected_objective_value(self) -> Array:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """Bounds for variables - free variables."""
        return None



"""
Problem OPTMASS: Optimal control of a particle on a frictionless plane.

A particle of unit mass moves on a frictionless plane under the action of a
controlling force whose magnitude may not exceed unity. At time=0, the particle
moves through the origin in the direction of the positive x-axis with speed SPEED.
The cost function maximizes final distance from origin while minimizing final speed.

Variables (for j=1,2 and i=0 to N+1):
- X(j,i): position coordinates at time i/N
- V(j,i): velocity components at time i/N
- F(j,i): force components at time i/N (only for i=0 to N)

Source:
M. Gawande and J. Dunn,
"A Projected Newton Method in a Cartesian Product of Balls",
JOTA 59(1): 59-69, 1988.

SIF input: Ph. Toint, June 1990.

classification QQR2-AN-V-V
"""

import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class OPTMASS(AbstractConstrainedMinimisation):
    """OPTMASS: Optimal control of particle on frictionless plane.

    Problem size depends on discretization parameter N:
    - Variables: 6*(N+2) - 2 total (X, V, F for each time step)
    - Constraints: 4*(N+1) equality + (N+1) inequality constraints

    Variable ordering (from SIF file):
    For i=0 to N: X(1,i), V(1,i), F(1,i), X(2,i), V(2,i), F(2,i)
    For i=N+1: X(1,N+1), V(1,N+1), X(2,N+1), V(2,N+1)
    """

    # Problem parameters
    N: int = 5000  # Default from SIF file (n=30010)
    SPEED: float = 0.01  # Initial speed
    PEN: float = 0.335  # Penalty parameter for final speed

    @property
    def name(self) -> str:
        return "OPTMASS"

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def n(self) -> int:
        # For i=0 to N: 6 variables each (X1, X2, V1, V2, F1, F2)
        # For i=N+1: 4 variables (X1, X2, V1, V2)
        return 6 * (self.N + 1) + 4

    @property
    def m(self) -> int:
        # A(j,i): 2*(N+1) equations
        # B(j,i): 2*(N+1) equations
        # C(i): (N+1) inequalities
        return 4 * (self.N + 1) + (self.N + 1)

    def _get_variable(self, var_type, j, i):
        """Get index of variable X, V, or F at position (j,i).

        var_type: 'X', 'V', or 'F'
        j: component (1 or 2)
        i: time step (0 to N+1)
        """
        if i <= self.N:
            # For i=0 to N: each timestep has 6 variables
            # Order: X(1,i), V(1,i), F(1,i), X(2,i), V(2,i), F(2,i)
            base = 6 * i
            if j == 1:
                if var_type == "X":
                    return base + 0
                elif var_type == "V":
                    return base + 1
                else:  # F
                    return base + 2
            else:  # j == 2
                if var_type == "X":
                    return base + 3
                elif var_type == "V":
                    return base + 4
                else:  # F
                    return base + 5
        else:  # i == N+1
            # For i=N+1: 4 variables
            # Order: X(1,N+1), V(1,N+1), X(2,N+1), V(2,N+1)
            base = 6 * (self.N + 1)
            if j == 1:
                if var_type == "X":
                    return base + 0
                else:  # V
                    return base + 1
            else:  # j == 2
                if var_type == "X":
                    return base + 2
                else:  # V
                    return base + 3

    @property
    def y0(self):
        # Initial point: all zeros except V(1,0) = SPEED
        y0_vals = jnp.zeros(self.n)
        # V(1,0) is at index 2 (after X(1,0), X(2,0))
        v1_0_idx = self._get_variable("V", 1, 0)
        y0_vals = y0_vals.at[v1_0_idx].set(self.SPEED)
        return y0_vals

    @property
    def args(self):
        return None

    def objective(self, y, args):
        """Maximize distance, minimize speed:
        -X(1,N+1)^2 - X(2,N+1)^2 + PEN*(V(1,N+1)^2 + V(2,N+1)^2).
        """
        N = self.N

        # Final position and velocity
        x1_final = y[self._get_variable("X", 1, N + 1)]
        x2_final = y[self._get_variable("X", 2, N + 1)]
        v1_final = y[self._get_variable("V", 1, N + 1)]
        v2_final = y[self._get_variable("V", 2, N + 1)]

        # Objective: minimize negative distance + penalty on speed
        return -(x1_final**2 + x2_final**2) + self.PEN * (v1_final**2 + v2_final**2)

    def constraint(self, y):
        """State equations and force magnitude constraints (vectorized).

        Equality constraints (SIF order: A(1,1), B(1,1), A(2,1), B(2,1), ...):
        A(j,i): X(j,i) - X(j,i-1) - V(j,i-1)/N - F(j,i-1)/(2N^2) = 0
        B(j,i): V(j,i) - V(j,i-1) - F(j,i-1)/N = 0

        Inequality constraints:
        C(i): F(1,i)^2 + F(2,i)^2 <= 1
        """
        N = self.N
        h = 1.0 / N  # time step
        h2_half = h * h / 2.0

        # Extract all variables efficiently
        # For i=0 to N: reshape to (N+1, 6) where each row is [X1, V1, F1, X2, V2, F2]
        vars_main = y[: 6 * (N + 1)].reshape((N + 1, 6))
        X_main = vars_main[:, [0, 3]]  # Shape: (N+1, 2) - X(j,i) for j=1,2 and i=0 to N
        V_main = vars_main[:, [1, 4]]  # Shape: (N+1, 2) - V(j,i) for j=1,2 and i=0 to N
        F_main = vars_main[:, [2, 5]]  # Shape: (N+1, 2) - F(j,i) for j=1,2 and i=0 to N

        # For i=N+1: X(j,N+1), V(j,N+1) for j=1,2
        vars_final = y[6 * (N + 1) :].reshape((1, 4))
        X_final = vars_final[:, [0, 2]]  # Shape: (1, 2) - X(j,N+1) for j=1,2
        V_final = vars_final[:, [1, 3]]  # Shape: (1, 2) - V(j,N+1) for j=1,2

        # Combine position and velocity arrays for full time series
        X_all = jnp.vstack([X_main, X_final])  # Shape: (N+2, 2)
        V_all = jnp.vstack([V_main, V_final])  # Shape: (N+2, 2)

        # State equations for i=1 to N+1 (vectorized)
        i_indices = jnp.arange(1, N + 2)  # i from 1 to N+1

        # A(j,i): position equations for all j and i
        # X(j,i) - X(j,i-1) - h*V(j,i-1) - h2_half*F(j,i-1) = 0
        A_eq = (
            X_all[i_indices]
            - X_all[i_indices - 1]
            - h * V_all[i_indices - 1]
            - h2_half * F_main[i_indices - 1]
        )

        # B(j,i): velocity equations for all j and i
        # V(j,i) - V(j,i-1) - h*F(j,i-1) = 0
        B_eq = V_all[i_indices] - V_all[i_indices - 1] - h * F_main[i_indices - 1]

        # Combine constraints in SIF order: A(1,i), B(1,i), A(2,i), B(2,i) for each i
        equality_constraints = jnp.concatenate(
            [
                jnp.stack(
                    [A_eq[:, 0], B_eq[:, 0], A_eq[:, 1], B_eq[:, 1]], axis=1
                ).flatten()
            ]
        )

        # Force magnitude constraints for i=0 to N (vectorized)
        # F(1,i)^2 + F(2,i)^2 <= 1 converted to pycutest form:
        # F(1,i)^2 + F(2,i)^2 - 1 <= 0
        F_squared_sum = jnp.sum(F_main**2, axis=1)  # Shape: (N+1,)
        inequality_constraints = F_squared_sum - 1.0

        return equality_constraints, inequality_constraints

    @property
    def bounds(self):
        """Variable bounds from SIF file.

        Fixed values:
        - X(1,0) = 0, X(2,0) = 0 (start at origin)
        - V(1,0) = SPEED, V(2,0) = 0 (initial velocity)
        All other variables are free.
        """
        lower = jnp.full(self.n, -jnp.inf)
        upper = jnp.full(self.n, jnp.inf)

        # X(1,0) = 0.0
        idx = self._get_variable("X", 1, 0)
        lower = lower.at[idx].set(0.0)
        upper = upper.at[idx].set(0.0)

        # X(2,0) = 0.0
        idx = self._get_variable("X", 2, 0)
        lower = lower.at[idx].set(0.0)
        upper = upper.at[idx].set(0.0)

        # V(1,0) = SPEED
        idx = self._get_variable("V", 1, 0)
        lower = lower.at[idx].set(self.SPEED)
        upper = upper.at[idx].set(self.SPEED)

        # V(2,0) = 0.0
        idx = self._get_variable("V", 2, 0)
        lower = lower.at[idx].set(0.0)
        upper = upper.at[idx].set(0.0)

        return lower, upper

    @property
    def expected_result(self):
        """Solution not provided in SIF file."""
        return None

    @property
    def expected_objective_value(self):
        """Solution not provided in SIF file."""
        return None



import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractUnconstrainedMinimisation


class LUKSAN21LS(AbstractUnconstrainedMinimisation):
    """Problem 21 (modified discrete boundary value) from Luksan.

    This is a least squares problem from the paper:
    L. Luksan
    "Hybrid methods in large sparse nonlinear least squares"
    J. Optimization Theory & Applications 89(3) 575-595 (1996)

    SIF input: Nick Gould, June 2017.

    least-squares version

    classification SUR2-AN-V-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})
    n: int = 100  # Number of variables (default from SIF)

    @property
    def y0(self) -> Array:
        """Initial guess: x(i) = i*h*(i*h - 1) where h = 1/(n+1)."""
        h = 1.0 / (self.n + 1)
        i_vals = jnp.arange(1, self.n + 1, dtype=jnp.float64)
        ih = i_vals * h
        return ih * (ih - 1.0)

    @property
    def args(self):
        """No additional arguments."""
        return None

    def objective(self, y: Array, args) -> Array:
        """Compute the least squares objective function.

        The objective is the sum of squares of M = N residuals.
        The residuals come from a discretized boundary value problem:
        - E(1): 2*x(1) - x(2) + h^2/2 * (x(1) + h + 1)^3 + 1
        - E(i): 2*x(i) - x(i-1) - x(i+1) + h^2/2 * (x(i) + h*i + 1)^3 + 1  (i=2,...,N-1)
        - E(N): 2*x(N) - x(N-1) + h^2/2 * (x(N) + h*N + 1)^3 + 1
        """
        del args  # Not used

        x = y
        n = self.n

        # Parameters
        h = 1.0 / (n + 1)
        h2 = h * h
        h2_half = 0.5 * h2

        # Create index arrays for vectorized computation
        i_vals = jnp.arange(
            1, n + 1, dtype=x.dtype
        )  # 1 to n (1-indexed as in original)
        hi_vals = h * i_vals  # h * i for all i

        # Compute cubic terms for all equations: (x(i) + h*i + 1)^3
        xhip1_vals = x + hi_vals + 1.0  # x(i) + h*i + 1 for all i
        cubic_terms = h2_half * (xhip1_vals**3)  # h^2/2 * (x(i) + h*i + 1)^3

        # Handle the three cases using vectorized operations
        # Case 1: First equation (i=1): 2*x(1) - x(2) + cubic_term + 1
        # Case 2: Middle eqs (2 <= i <= n-1): 2*x(i) - x(i-1) - x(i+1) + cubic + 1
        # Case 3: Last equation (i=n): 2*x(n) - x(n-1) + cubic_term + 1

        # Initialize residuals with 2*x(i) + cubic_term + 1 for all i
        residuals = 2.0 * x + cubic_terms + 1.0

        # Subtract x(i+1) from equations 1 to n-1
        residuals = residuals.at[:-1].add(-x[1:])

        # Subtract x(i-1) from equations 2 to n
        residuals = residuals.at[1:].add(-x[:-1])

        # Sum of squares (L2 group type in SIF)
        return jnp.sum(residuals**2)

    @property
    def expected_result(self) -> Array | None:
        """Expected optimal solution."""
        # No specific expected result provided in SIF
        return None

    @property
    def expected_objective_value(self) -> Array | None:
        """Expected objective value."""
        # For least squares, the expected value is 0
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class HS30(AbstractConstrainedMinimisation):
    """Problem 30 from the Hock-Schittkowski test collection.

    A 3-variable quadratic function with one inequality constraint and bounds.

    f(x) = x₁² + x₂² + x₃²

    Subject to: x₁² + x₂² - 1 ≥ 0
                1 ≤ x₁ ≤ 10
                -10 ≤ x₂ ≤ 10
                -10 ≤ x₃ ≤ 10

    Source: problem 30 in
    W. Hock and K. Schittkowski,
    "Test examples for nonlinear programming codes",
    Lectures Notes in Economics and Mathematical Systems 187, Springer
    Verlag, Heidelberg, 1981.

    Also cited: Betts [8]

    Classification: QQR-T1-8
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y, args):
        x1, x2, x3 = y
        return x1**2 + x2**2 + x3**2

    @property
    def y0(self):
        return jnp.array([1.0, 1.0, 1.0])

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        return jnp.array([1.0, 0.0, 0.0])

    @property
    def expected_objective_value(self):
        return jnp.array(1.0)

    @property
    def bounds(self):
        return (jnp.array([1.0, -10.0, -10.0]), jnp.array([10.0, 10.0, 10.0]))

    def constraint(self, y):
        x1, x2, x3 = y
        # Inequality constraint: x₁² + x₂² - 1 ≥ 0
        inequality_constraint = jnp.array([x1**2 + x2**2 - 1.0])
        return None, inequality_constraint



import jax.numpy as jnp

from ..._misc import inexact_asarray
from ..._problem import AbstractNonlinearEquations


# TODO: Human review needed - pycutest returns 0.0 at starting point
# Similar to VANDERM1 issue
# TODO: Human review needed - constraint values don't match pycutest
class VANDERM4(AbstractNonlinearEquations):
    """VANDERM4 problem - Vandermonde matrix nonlinear equation system.

    A nonlinear equations problem, subject to monotonicity constraints.
    The Jacobian is a dense Vandermonde matrix.

    Problems VANDERM1, VANDERM2, VANDERM3 and VANDERM4 differ by the rhs
    of the equation. They are increasingly degenerate.

    The problem is non-convex.

    Source:
    A. Neumaier, private communication, 1991.

    SIF input: Ph. L. Toint, May 1993.

    Classification: NOR2-AN-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 100  # Number of variables (default 100)

    def num_residuals(self):
        """Number of residuals."""
        # n residual equations (Vandermonde equations)
        return self.n

    def residual(self, y, args):
        """Compute the residuals."""
        del args
        n = self.n
        x = y

        # Define the right-hand-side for residual equations
        # al[i] = 1/2^(i-1) for i = 1, ..., n
        al = 1.0 / (2.0 ** inexact_asarray(jnp.arange(n)))

        # Compute A values
        a = jnp.zeros(n)
        a = a.at[0].set(jnp.sum(al))

        # For k >= 2, use log/exp to compute al^k to avoid numerical issues
        for k in range(2, n + 1):
            # al^k = exp(k * log(al))
            log_al = jnp.log(al)
            al_k = jnp.exp(k * log_al)
            a = a.at[k - 1].set(jnp.sum(al_k))

        # Residual equations: the Vandermonde equations
        residuals = jnp.zeros(n)

        # First equation: sum(x_i) = a[0]
        # Note: pycutest appears to use a[k] - sum(x^k) convention
        residuals = residuals.at[0].set(a[0] - jnp.sum(x))

        # Remaining equations: sum(x_i^k) = a[k-1]
        for k in range(2, n + 1):
            residuals = residuals.at[k - 1].set(a[k - 1] - jnp.sum(x**k))

        return residuals

    @property
    def y0(self):
        """Initial guess."""
        n = self.n
        # Initial point: x[i] = (i-1)/n
        return inexact_asarray(jnp.arange(n)) * n

    @property
    def args(self):
        """Additional arguments (none for this problem)."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution (not provided in SIF)."""
        return None

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        return jnp.array(0.0)

    def constraint(self, y):
        """Compute the constraints (both equality and inequality)."""
        # Equality constraints: the residuals
        equalities = self.residual(y, self.args)

        # Inequality constraints: monotonicity x[i] >= x[i-1] for i = 2, ..., n
        # Rewritten as x[i] - x[i-1] >= 0
        x = y
        n = self.n
        inequalities = x[1:n] - x[0 : n - 1]

        return equalities, inequalities

    @property
    def bounds(self) -> tuple[jnp.ndarray, jnp.ndarray] | None:
        """No bounds for this problem."""
        return None



import jax
import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractNonlinearEquations


class CHEBYQADNE(AbstractNonlinearEquations):
    """
    The Chebyquad problem using the exact formula for the
    shifted chebyshev polynomials.
    The Hessian is full.

    Source: problem 35 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    See also Buckley#133 (p. 44).
    SIF input: Nick Gould, March 1990.
    Bound-constrained nonlinear equations version: Nick Gould, June 2019.

    classification NOR2-AN-V-V
    """

    n: int = 100  # Number of variables (also M)
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def chebypol(self, x, i):
        """Evaluate the i-th shifted Chebyshev polynomial at x"""
        dif = 2.0 * x - 1.0
        # Use arccos to compute Chebyshev polynomial
        # Convert i to same dtype as x to avoid promotion issues
        i_float = jnp.asarray(i, dtype=x.dtype)
        acosx = i_float * jnp.arccos(dif)
        return jnp.cos(acosx)

    def residual(self, y: Array, args) -> Array:
        """Compute the residuals"""
        n = self.n
        m = n  # M = N in this problem

        # Create array of indices 1 to m as floats to avoid dtype issues
        indices = jnp.arange(1, m + 1, dtype=y.dtype)

        # Vectorized computation of Chebyshev polynomials for all i and all y[j]
        # First, vmap over y to compute chebypol for each y[j] and all i values
        cheby_vals_per_y = jax.vmap(
            lambda yj: jax.vmap(lambda i: self.chebypol(yj, i))(indices)
        )(y)

        # Sum over all y values and divide by n
        residuals = jnp.sum(cheby_vals_per_y, axis=0) / n

        # Subtract constants for even indices
        is_even = (indices % 2) == 0
        i_squared = indices * indices
        constants = jnp.where(is_even, -1.0 / (i_squared - 1.0), 0.0)
        residuals = residuals - constants

        return residuals

    @property
    def y0(self) -> Array:
        """Initial guess for the optimization problem."""
        # Starting point: j/(n+1) for j = 1, ..., n
        n = self.n
        return jnp.arange(1, n + 1, dtype=jnp.float64) / (n + 1)

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> None:
        """Expected result of the optimization problem."""
        # No exact solution provided in SIF file
        return None

    @property
    def expected_objective_value(self) -> Array:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """Bounds on variables: all in [0, 1]"""
        n = self.n
        lower = jnp.zeros(n, dtype=jnp.float64)
        upper = jnp.ones(n, dtype=jnp.float64)
        return lower, upper



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class ELATTAR(AbstractConstrainedMinimisation):
    """A nonlinear minmax problem in six variables.

    The problem is nonconvex and has several local minima.

    Source:
    R.A. El-Attar, M. Vidyasagar and S.R.K. Dutta,
    "An algorithm for l_1-approximation",
    SINUM 16, pp.70-86, 1979.

    SIF input: Ph. Toint, Nov 1993.

    Classification: LOR2-AN-7-102
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def n(self):
        """Number of variables."""
        return 7  # 6 x variables + 1 u variable

    def objective(self, y, args):
        """Compute the objective function: minimize u."""
        del args
        # The objective is simply u (the 7th variable)
        return y[6]

    def constraint(self, y):
        """Returns the inequality constraints.

        Constraints are:
        u >= f_i(x) - y_i for all i (51 constraints)
        u >= -(f_i(x) - y_i) for all i (51 constraints)

        Rewritten as:
        f_i(x) - y_i - u <= 0
        -(f_i(x) - y_i) - u <= 0
        """
        # Extract variables
        x = y[:6]
        u = y[6]

        # Generate data points
        t_values = jnp.arange(51, dtype=jnp.float64) * 0.1

        # Compute y_i values
        y_i = (
            0.5 * jnp.exp(t_values)
            - jnp.exp(-2.0 * t_values)
            + 0.5 * jnp.exp(-3.0 * t_values)
            + 1.5 * jnp.exp(-1.5 * t_values) * jnp.sin(7.0 * t_values)
            + jnp.exp(-2.5 * t_values) * jnp.sin(5.0 * t_values)
        )

        # Compute f_i(x) = x1 * exp(-x2*t) * cos(x3*t + x4) + x5 * exp(-x6*t)
        # First element type: x1 * exp(-x2*t) * cos(x3*t + x4)
        a1 = -x[1] * t_values  # -x2 * t
        b1 = x[2] * t_values + x[3]  # x3 * t + x4
        et1 = x[0] * jnp.exp(a1) * jnp.cos(b1)

        # Second element type: x5 * exp(-x6*t)
        a2 = -x[5] * t_values  # -x6 * t
        et2 = x[4] * jnp.exp(a2)

        # f_i(x) = et1 + et2
        f_i = et1 + et2

        # Residuals
        residuals = f_i - y_i

        # Inequality constraints
        # The SIF file has F(I) and MF(I) groups
        # F(I): u - (f_i - y_i) >= 0  => f_i - y_i - u <= 0
        # MF(I): u - (-(f_i - y_i)) >= 0  => -(f_i - y_i) - u <= 0

        # Interleave the constraints as they appear in the SIF file
        constraints = []
        for i in range(51):
            constraints.append(residuals[i] - u)  # F(I)
            constraints.append(-residuals[i] - u)  # MF(I)

        # Return only inequality constraints
        return None, jnp.array(constraints)

    @property
    def y0(self):
        """Initial guess."""
        # From SIF file, the more interesting starting point
        return jnp.array([-2.0, -2.0, 7.0, 0.0, -2.0, 1.0, 0.0])

    @property
    def args(self):
        """No additional arguments."""
        return None

    @property
    def bounds(self):
        """No bounds on variables."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution (approximate)."""
        # The SIF file doesn't provide the solution
        return jnp.zeros(7)  # Placeholder

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        # From SIF file: either 0.1427066255 or 74.206179244
        # The smaller value seems more likely for a well-solved problem
        return jnp.array(0.1427066255)



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class LOGHAIRY(AbstractUnconstrainedMinimisation):
    """
    LOGHAIRY problem.

    A more difficult variant of the HAIRY problem in two variables.
    It is defined by a logarithmic transformation of the HAIRY surface,
    which is defined by this function has a large number of relatively
    sharp hills between which a valley leads to the minimizer.
    This problem contains a large number of saddle points.

    The problem is nonconvex.

    Source:
    Ph. Toint, private communication,

    SIF input: Ph. Toint, April 1997.

    classification OUR2-AN-2-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y, args):
        del args
        x1, x2 = y

        # Parameters
        hlength = 30.0
        cslope = 100.0

        # Element parameters
        dens = 7.0
        smooth = 0.01

        # FUR element
        dv1 = dens * x1
        dv2 = dens * x2
        s1sq = jnp.sin(dv1) ** 2
        c2sq = jnp.cos(dv2) ** 2
        fur_value = s1sq * c2sq

        # DCUP element (2D diagonal cup)
        v = x1 - x2
        vsq = v * v
        arg = smooth + vsq
        dcup_value = jnp.sqrt(arg)

        # 1CUP element (1D cup)
        v_1d = x1
        vsq_1d = v_1d * v_1d
        arg_1d = smooth + vsq_1d
        cup1_value = jnp.sqrt(arg_1d)

        # Group contributions with weights
        group_sum = hlength * fur_value + cslope * dcup_value + cslope * cup1_value

        # LOG group type transformation
        s = 100.0
        return jnp.log((s + group_sum) / s)

    @property
    def y0(self):
        return jnp.array([-500.0, -700.0])

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # Solution not provided in SIF file
        return None

    @property
    def expected_objective_value(self):
        # Solution value given in SIF file
        return jnp.array(0.1823216)



import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractConstrainedMinimisation


# TODO: Human review needed
# Attempts made: Multiple attempts to fix data generation and constraint formulation
# Suspected issues: Complex differences in objective/constraint formulation vs pycutest
# Resources needed: Deep understanding of how pycutest handles orthogonal regression
class ORTHREGA(AbstractConstrainedMinimisation):
    """
    An orthogonal regression problem.

    The problem is to fit (orthogonally) an ellipse to a set of points
    in the plane.

    Source:
    M. Gulliksson,
    "Algorithms for nonlinear Least-squares with Applications to
    Orthogonal Regression",
    UMINF-178.90, University of Umea, Sweden, 1990.

    SIF input: Ph. Toint, June 1990.

    classification QQR2-AN-V-V

    Number of levels in the generation of the data points
    ( number of data points =     4**LEVELS
      number of variables   = 2 * 4**LEVELS + 5
      number of constraints =     4**LEVELS         )
    Default LEVELS = 3 (original value)
    """

    levels: int = 6  # Number of levels in data generation (default from SIF)
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def __init__(self, levels: int = 6):
        self.levels = levels

    @property
    def n(self) -> int:
        """Total number of variables: 2 * 4**levels + 5."""
        npts = 4**self.levels
        return 2 * npts + 5

    @property
    def m(self) -> int:
        """Number of constraints: 4**levels equality constraints."""
        return 4**self.levels

    def _generate_data_points(self):
        """Generate the data points for the ellipse fitting."""
        # Parameters for data generation
        a = 9.0
        b = 6.0
        cx = 0.5
        cy = 0.5
        pi = jnp.pi

        # Initialize with center point
        xd = [cx]
        yd = [cy]

        # Generate points level by level
        for i in range(1, self.levels + 1):
            # Store current points
            xz = xd.copy()
            yz = yd.copy()

            # Clear for new points
            xd = []
            yd = []

            # Generate 4 new points for each existing point
            for j in range(len(xz)):
                # Add 4 points around (xz[j], yz[j])
                xd.append(xz[j] + a)
                yd.append(yz[j] + a)

                xd.append(xz[j] + b)
                yd.append(yz[j] - b)

                xd.append(xz[j] - a)
                yd.append(yz[j] - a)

                xd.append(xz[j] - b)
                yd.append(yz[j] + b)

            # Update a and b for next level (divide by pi)
            a = a / pi
            b = b / pi

        return jnp.array(xd, dtype=jnp.float64), jnp.array(yd, dtype=jnp.float64)

    def starting_point(self) -> Array:
        """Return the starting point for the problem."""
        npts = 4**self.levels
        y = jnp.zeros(self.n, dtype=jnp.float64)

        # Ellipse parameters
        y = y.at[0].set(1.0)  # h11
        y = y.at[1].set(0.0)  # h12
        y = y.at[2].set(1.0)  # h22
        y = y.at[3].set(0.0)  # g1
        y = y.at[4].set(0.0)  # g2

        # X and Y projections initialized to data points
        xd, yd = self._generate_data_points()
        y = y.at[5 : 5 + npts].set(xd)
        y = y.at[5 + npts :].set(yd)

        return y

    def objective(self, y: Array, args) -> Array:
        """Compute the objective function."""
        npts = 4**self.levels

        # Extract variables
        x = y[5 : 5 + npts]
        y_points = y[5 + npts :]

        # Get data points
        xd, yd = self._generate_data_points()

        # Objective: sum of squared distances to data points
        obj = jnp.sum((x - xd) ** 2 + (y_points - yd) ** 2)

        return obj

    def constraint(self, y: Array):
        """Compute the equality and inequality constraints."""
        npts = 4**self.levels

        # Extract variables
        h11 = y[0]
        h12 = y[1]
        h22 = y[2]
        g1 = y[3]
        g2 = y[4]
        x = y[5 : 5 + npts]
        y_points = y[5 + npts :]

        # Equality constraints
        # E(i): ellipse constraint equation
        eq_constraints = (
            h11 * x**2
            + 2.0 * h12 * x * y_points
            + h22 * y_points**2
            - 2.0 * g1 * x
            - 2.0 * g2 * y_points
            + 1.0
        )

        # No inequality constraints
        ineq_constraints = None

        return eq_constraints, ineq_constraints

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """All variables are free."""
        return None

    @property
    def y0(self) -> Array:
        """Initial guess for the optimization problem."""
        return self.starting_point()

    @property
    def args(self):
        """Additional arguments for the objective and constraint functions."""
        return None

    @property
    def expected_result(self) -> Array:
        """Expected result of the optimization problem."""
        # Not explicitly given in the SIF file
        return jnp.zeros(self.n, dtype=jnp.float64)

    @property
    def expected_objective_value(self) -> Array:
        """Expected value of the objective at the solution."""
        # From SIF file comments:
        # SOLTN(3) = 350.29936756
        # SOLTN(4) = 1414.0524915
        return jnp.array(350.29936756)



import jax
import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: This implementation requires human review and verification against
# another CUTEst interface
class CHWIRUT2LS(AbstractUnconstrainedMinimisation):
    """NIST Data fitting problem CHWIRUT2.

    The objective is to minimize a least-squares function with the model:
    y = exp[-b1*x]/(b2+b3*x) + e

    where b1, b2, b3 are the parameters to be determined and e represents
    the error term.

    Source: Problem from the NIST nonlinear regression test set
      http://www.itl.nist.gov/div898/strd/nls/nls_main.shtml

    Reference: Chwirut, D., NIST (197?).
      Ultrasonic Reference Block Study.

    SIF input: Nick Gould and Tyrone Rees, Oct 2015

    Classification: SUR2-MN-3-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 3  # Number of variables
    m: int = 54  # Number of data points

    def objective(self, y, args):
        del args
        b1, b2, b3 = y

        # X data points from the SIF file
        x_data = jnp.array(
            [
                0.5,
                1.0,
                1.75,
                3.75,
                5.75,
                0.875,
                2.25,
                3.25,
                5.25,
                0.75,
                1.75,
                2.75,
                4.75,
                0.625,
                1.25,
                2.25,
                4.25,
                0.5,
                3.0,
                0.75,
                3.0,
                1.5,
                6.0,
                3.0,
                6.0,
                1.5,
                3.0,
                0.5,
                2.0,
                4.0,
                0.75,
                2.0,
                5.0,
                0.75,
                2.25,
                3.75,
                5.75,
                3.0,
                0.75,
                2.5,
                4.0,
                0.75,
                2.5,
                4.0,
                0.75,
                2.5,
                4.0,
                0.5,
                6.0,
                3.0,
                0.5,
                2.75,
                0.5,
                1.75,
            ]
        )

        # Y data points from the SIF file
        y_data = jnp.array(
            [
                92.9,
                57.1,
                31.05,
                11.5875,
                8.025,
                63.6,
                21.4,
                14.25,
                8.475,
                63.8,
                26.8,
                16.4625,
                7.125,
                67.3,
                41.0,
                21.15,
                8.175,
                81.5,
                13.12,
                59.9,
                14.62,
                32.9,
                5.44,
                12.56,
                5.44,
                32.0,
                13.95,
                75.8,
                20.0,
                10.42,
                59.5,
                21.67,
                8.55,
                62.0,
                20.2,
                7.76,
                3.75,
                11.81,
                54.7,
                23.7,
                11.55,
                61.3,
                17.7,
                8.74,
                59.2,
                16.3,
                8.62,
                81.0,
                4.87,
                14.62,
                81.7,
                17.17,
                81.3,
                28.9,
            ]
        )

        # Define function to compute model value for a single x
        def compute_model(x):
            # Model: y = exp[-b1*x]/(b2+b3*x)
            numerator = jnp.exp(-b1 * x)
            denominator = b2 + b3 * x
            return numerator / denominator

        # Compute model predictions for all x values using vmap
        y_pred = jax.vmap(compute_model)(x_data)

        # Compute residuals and return sum of squares
        residuals = y_pred - y_data
        return jnp.sum(residuals**2)

    @property
    def y0(self):
        # Initial values from SIF file (START1)
        return jnp.array([0.1, 0.01, 0.02])

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # The optimal solution is not directly provided in the SIF file
        # Values from NIST: 0.167, 0.0052, 0.011
        return jnp.array([0.167, 0.0052, 0.011])

    @property
    def expected_objective_value(self):
        # Certified value from NIST: 513.05
        return jnp.array(513.05)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class COSHFUN(AbstractConstrainedMinimisation):
    """
    COSHFUN problem in CUTEst.

    A nonlinear minmax problem.

    Source:
    K. Jonasson and K. Madsen,
    "Corrected sequential linear programming for sparse
    minimax optimization", Technical report, Institute for Numerical
    Analysis, Technical U. of Denmark.

    classification LOR2-AN-V-V

    SIF input: Nick Gould, October 1992.
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Number of functions
    m: int = 2000  # Default from SIF, can be 3, 8, 14, 20, 200, 2000

    @property
    def n(self) -> int:
        """Number of variables (3 * m)."""
        return 3 * self.m

    @property
    def initial_guess(self):
        """Initial guess - all zeros (default when not specified in SIF)."""
        n = self.n
        y0 = jnp.zeros(n + 1, dtype=jnp.float64)  # n variables + f
        return y0

    @property
    def y0(self):
        """Initial guess."""
        return self.initial_guess

    @property
    def args(self):
        """Additional arguments (none for this problem)."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution (not provided in SIF)."""
        return None

    @property
    def expected_objective_value(self):
        """Expected optimal objective value (not provided in SIF)."""
        return None

    @property
    def bounds(self):
        """Variable bounds (unbounded)."""
        return None

    def objective(self, y, args):
        """Objective function - minimize f (the last variable)."""
        del args  # Unused
        return y[-1]

    def constraint(self, y):
        """Compute constraints."""
        # For a minmax problem, we have inequalities: c_i(x) <= f
        # Which we express as: c_i(x) - f <= 0
        inequalities = self.inequality_constraints(y)
        return None, inequalities

    def inequality_constraints(self, y):
        """Compute inequality constraints c_i(x) - f <= 0."""
        m = self.m
        n = self.n
        x = y[:-1]  # First n variables
        f = y[-1]  # Last variable is f

        # Vectorized indices for all constraints
        i_range = jnp.arange(m)

        # Indices for the nonlinear terms (vectorized)
        idx_sqr = 3 * i_range + 2  # X(3*i) in 1-based
        idx_cosh = 3 * i_range + 1  # X(3*i-1) in 1-based
        idx_prod1 = 3 * i_range  # X(3*i-2) in 1-based
        idx_prod2 = 3 * i_range + 2  # X(3*i) in 1-based

        # Nonlinear terms: x^2 + cosh(x) + 2*x*x*y (vectorized)
        c_vec = (
            x[idx_sqr] ** 2
            + jnp.cosh(x[idx_cosh])
            + 2.0 * x[idx_prod1] ** 2 * x[idx_prod2]
        )

        # Linear terms for the first constraint (C1)
        c_vec = c_vec.at[0].add(-2.0 * x[2] - x[5])

        # Vectorized linear terms for middle constraints
        # From SIF: C(I/3) has X(I-5), X(I), X(I+3) terms
        # where I goes from 6 to N-3 in steps of 3
        middle_indices = jnp.arange(1, m - 1)
        idx_base = 3 * (middle_indices + 1)  # This gives us 6, 9, 12, ...

        # Create mask for valid middle constraints
        valid_mask = idx_base < n - 2

        # Compute linear terms for all middle indices, using 0 for invalid ones
        linear_terms = jnp.where(
            valid_mask,
            x[jnp.minimum(idx_base - 5, n - 1)]
            - 2.0 * x[jnp.minimum(idx_base, n - 1)]
            - x[jnp.minimum(idx_base + 3, n - 1)],
            0.0,
        )

        # Add linear terms to middle constraints
        c_vec = c_vec.at[middle_indices].add(linear_terms)

        # Linear terms for the last constraint (C(M))
        c_vec = c_vec.at[m - 1].add(x[n - 6] - 2.0 * x[n - 1])

        # Return as c_i - f <= 0
        return c_vec - f



import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractNonlinearEquations


class GENROSENE(AbstractNonlinearEquations):
    """
    The generalized Rosenbrock function.
    This is a nonlinear equation variant of GENROSE

    Source: problem 5 in
    S. Nash,
    "Newton-type minimization via the Lanczos process",
    SIAM J. Num. Anal. 21, 1984, 770-788.

    SIF input: Nick Gould, Oct 1992.
              minor correction by Ph. Shott, Jan 1995.

    classification NOR2-AN-V-V
    """

    n: int = 1000  # Default value from SIF
    m: int = 1999  # 1 + 2*(n-1) equations
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def starting_point(self) -> Array:
        # From SIF: X(I) = I/(N+1)
        i_vals = jnp.arange(1, self.n + 1, dtype=jnp.float64)
        return i_vals / (self.n + 1)

    def num_residuals(self) -> int:
        return self.m

    def residual(self, y: Array, args) -> Array:
        """Compute the residuals of the generalized Rosenbrock problem"""
        x = y
        n = self.n

        # Initialize residuals - we have 1 + 2*(n-1) equations
        residuals = jnp.zeros(1 + 2 * (n - 1), dtype=jnp.float64)

        # OBJ equation: constant is -1.0 in SIF, but we return the negative
        residuals = residuals.at[0].set(1.0)

        # For each i from 2 to n, we have 2 equations
        for i in range(1, n):  # i = 2 to n in 1-based indexing
            # Q(i) equation with SCALE 0.1
            # Element MSQR: -V^2 where V = x(i-1)
            # SCALE parameter divides the residual
            residuals = residuals.at[1 + 2 * (i - 1)].set((x[i] - x[i - 1] ** 2) / 0.1)

            # L(i) equation: x(i) - 1.0 = 0
            residuals = residuals.at[1 + 2 * (i - 1) + 1].set(x[i] - 1.0)

        return residuals

    @property
    def y0(self) -> Array:
        """Initial guess for the optimization problem."""
        return self.starting_point()

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> Array:
        """Expected result of the optimization problem."""
        # Solution would be all ones
        return jnp.ones(self.n, dtype=jnp.float64)

    @property
    def expected_objective_value(self) -> Array:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """No bounds for this problem."""
        return None



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: This problem currently does not run. It should be debugged.
# TODO: This implementation requires human review and verification against
# another CUTEst interface
class FMINSURF(AbstractUnconstrainedMinimisation):
    """The FMINSURF function.

    The free boundary minimum surface problem.

    The problem comes from the discretization of the minimum surface
    problem on the unit square with "free boundary conditions"
    one must find the minumum surface over the unit square
    (which is clearly 1.0). Furthermore, the average distance of the surface
    from zero is also minimized.

    The Hessian is dense.

    The unit square is discretized into (p-1)**2 little squares. The
    heights of the considered surface above the corners of these little
    squares are the problem variables, There are p**2 of them.
    Given these heights, the area above a little square is
    approximated by the
      S(i,j) = sqrt( 1 + 0.5(p-1)**2 ( a(i,j) + b(i,j) ) ) / (p-1)**2
    where
      a(i,j) = x(i,j) - x(i+1,j+1)
    and
      b(i,j) = x(i+1,j) - x(i,j+1)

    Source: setting the boundary free in
    A Griewank and Ph. Toint,
    "Partitioned variable metric updates for large structured
    optimization problems",
    Numerische Mathematik 39:429-448, 1982.

    SIF input: Ph. Toint, November 1991.
    Classification: OUR2-MY-V-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    p: int = 75  # Number of points on one side of the unit square
    h00: float = 1.0  # Boundary condition parameters
    slopej: float = 4.0  # Boundary condition parameters
    slopei: float = 8.0  # Boundary condition parameters

    @property
    def n(self):
        # Total number of variables: p^2
        return self.p * self.p

    def objective(self, y, args):
        del args
        p = self.p

        # Reshape the variables into a 2D grid
        x = y.reshape((p, p))

        # From AMPL file - the objective is:
        # sum {i in 1..p-1, j in 1..p-1}
        # sqrt(0.5*(p-1)^2*((x[i,j]-x[i+1,j+1])^2+(x[i+1,j]-x[i,j+1])^2)+1.0)/scale +
        # (sum {j in 1..p, i in 1..p} x[i,j])^2/p^4
        # where scale = (p-1)^2

        # First part: sum of area elements (vectorized)
        scale = (p - 1) ** 2

        # Vectorized computation of all a_ij and b_ij
        a_vals = x[:-1, :-1] - x[1:, 1:]  # x[i,j] - x[i+1,j+1]
        b_vals = x[1:, :-1] - x[:-1, 1:]  # x[i+1,j] - x[i,j+1]

        # Compute all area elements at once
        area_elements = jnp.sqrt(0.5 * scale * (a_vals**2 + b_vals**2) + 1.0) / scale
        area_sum = jnp.sum(area_elements)

        # Second part: (sum of all x values)^2 / p^4
        total_sum = jnp.sum(x)
        penalty = (total_sum**2) / (p**4)

        return area_sum + penalty

    @property
    def y0(self):
        # Initialize with zeros, then set boundary values
        p = self.p
        x = jnp.zeros((p, p))

        # Constants from AMPL file
        h00 = self.h00
        wtoe = self.slopej / (p - 1)
        ston = self.slopei / (p - 1)
        h01 = h00 + self.slopej
        h10 = h00 + self.slopei

        # From AMPL file (converting 1-based to 0-based indexing correctly):
        # let {j in 1..p} x[1,j] := (j-1)*wtoe+h00;
        # -> x[0,j-1] := (j-1)*wtoe+h00
        # let {j in 1..p} x[p,j] := (j-1)*wtoe+h10;
        # -> x[p-1,j-1] := (j-1)*wtoe+h10
        # let {i in 2..p-1} x[i,p] := (i-1)*ston+h00;
        # -> x[i-1,p-1] := (i-1)*ston+h00
        # let {i in 2..p-1} x[i,1] := (i-1)*ston+h01;
        # -> x[i-1,0] := (i-1)*ston+h01

        # Bottom edge (i=1 in AMPL = i=0 in 0-based): x[0,j] := (j-1)*wtoe+h00
        j_vals = jnp.arange(p)
        x = x.at[0, :].set((j_vals) * wtoe + h00)

        # Top edge (i=p in AMPL = i=p-1 in 0-based): x[p-1,j] := (j-1)*wtoe+h10
        x = x.at[p - 1, :].set((j_vals) * wtoe + h10)

        # Left edge (j=1 in AMPL = j=0 in 0-based):
        # x[i,0] := (i-1)*ston+h01 for i=2..p-1
        i_vals = jnp.arange(1, p - 1)
        x = x.at[i_vals, 0].set((i_vals) * ston + h01)

        # Right edge (j=p in AMPL = j=p-1 in 0-based):
        # x[i,p-1] := (i-1)*ston+h00 for i=2..p-1
        x = x.at[i_vals, p - 1].set((i_vals) * ston + h00)

        # Interior points: 0.0 (already initialized)

        # Flatten the 2D grid to 1D vector
        return x.flatten()

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        return None

    @property
    def expected_objective_value(self):
        return None



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class POLAK5(AbstractConstrainedMinimisation):
    """
    POLAK5 problem.

    A nonlinear minmax problem in two variables.

    Source:
    E. Polak, D.H. Mayne and J.E. Higgins,
    "Superlinearly convergent algorithm for min-max problems"
    JOTA 69, pp. 407-439, 1991.

    SIF input: Ph. Toint, Nov 1993.

    classification  LOR2-AN-3-2
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y, args):
        del args
        x1, x2, u = y
        # Linear objective: minimize u
        return u

    @property
    def y0(self):
        # Starting point
        return jnp.array([0.1, 0.1, 0.0])

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # Solution not provided in detail in SIF file
        return None

    @property
    def expected_objective_value(self):
        # Solution value provided as 50.0
        return jnp.array(50.0)

    def num_variables(self):
        return 3

    @property
    def bounds(self):
        # All variables are free (unbounded)
        return None

    def constraint(self, y):
        x1, x2, u = y

        # Element functions
        # E1: (x1 - x2^4 - 1)^2 with S = -1
        # E2: (x1 - x2^4 + 1)^2 with S = 1
        e1 = (x1 - x2**4 - 1.0) ** 2
        e2 = (x1 - x2**4 + 1.0) ** 2

        # X1SQ: x1^2
        x1sq = x1**2

        # Inequality constraints:
        # From SIF: F1 and F2 are L-type with coefficient -1.0 on U
        # F1: -u + 3*x1^2 + 50*e1 <= 0
        # F2: -u + 3*x1^2 + 50*e2 <= 0
        # pycutest returns raw values
        ineq_constraint = jnp.array(
            [3.0 * x1sq + 50.0 * e1 - u, 3.0 * x1sq + 50.0 * e2 - u]
        )

        # Equality constraints: none
        eq_constraint = None

        return eq_constraint, ineq_constraint



import jax.numpy as jnp

from ..._misc import inexact_asarray
from ..._problem import AbstractNonlinearEquations


class EIGENA(AbstractNonlinearEquations):
    """Solving symmetric eigenvalue problems as systems of nonlinear equations.

    The problem is, given a symmetric matrix A, to find an orthogonal
    matrix Q and diagonal matrix D such that A = Q(T) D Q.

    Example A: a diagonal matrix with eigenvalues 1, ..., N.

    Source: An idea by Nick Gould

    Nonlinear equations version.

    SIF input: Nick Gould, Nov 1992.

    Classification: NOR2-AN-V-V

    TODO: Human review needed

    Attempts made:
    1. Fixed initial values to match pycutest pattern
       (D[0]=D[1]=1.0, rest 0; Q sparse pattern)
    2. Added bounds property (lower=0 for all variables,
       interpreting SIF "nonnegative eigenvalues")
    3. Tested multiple constraint orderings
       (eigen-first vs ortho-first, interleaved, SIF loop order)
    4. Special handling for diagonal constraints with unused eigenvalues

    Current status: ~18/21 tests passing, systematic 50.0 constraint
    value discrepancies remain

    Suspected issues:
    - pycutest may interpret SIF constraint formulation differently
      than mathematical A = Q^T D Q
    - Constraint values differ by exactly eigenvalue magnitudes
      (50.0, 49.0, etc.)
    - Issue appears at diagonal elements of eigen-equation constraints

    Resources needed:
    - Deep investigation into pycutest's SIF constraint parsing
    - Understanding of how pycutest handles GROUPS and CONSTANTS sections
    - Possible consultation with pycutest maintainers
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Problem dimension
    N: int = 50  # Default from SIF file

    @property
    def n(self):
        """Number of variables: N eigenvalues + N*N eigenvector components."""
        return self.N + self.N * self.N

    def num_residuals(self):
        """Number of residuals: N(N+1)/2 eigen-eqs + N(N+1)/2 orthogonality eqs."""
        return self.N * (self.N + 1)  # Total of both sets of equations

    def residual(self, y, args):
        """Compute the residuals for eigenvalue problem.

        Residuals are:
        1. Eigen-equations: Q^T D Q - A = 0
        2. Orthogonality equations: Q^T Q - I = 0
        """
        del args

        # Extract eigenvalues and eigenvectors
        d = y[: self.N]  # Eigenvalues D(j)
        q = y[self.N :].reshape(self.N, self.N)  # Eigenvectors Q(i,j)

        # Define matrix A: diagonal with eigenvalues 1, ..., N
        a = inexact_asarray(jnp.diag(jnp.arange(1, self.N + 1)))

        # Vectorized computation
        # Compute Q^T D Q
        qtdq = q.T @ jnp.diag(d) @ q

        # Compute Q^T Q
        qtq = q.T @ q

        # Identity matrix
        eye = jnp.eye(self.N)

        residuals = []

        # Use the most successful ordering: eigen constraints first, ortho second
        # All eigen equations E(I,J) first
        for j in range(self.N):
            for i in range(j + 1):
                residuals.append(qtdq[i, j] - a[i, j])

        # All orthogonality equations O(I,J) second
        for j in range(self.N):
            for i in range(j + 1):
                residuals.append(qtq[i, j] - eye[i, j])

        return inexact_asarray(jnp.array(residuals))

    @property
    def y0(self):
        """Initial guess."""
        y0 = jnp.zeros(self.n)

        # From pycutest behavior (differs from SIF file):
        # Set D[0] and D[1] to 1.0
        y0 = y0.at[0].set(1.0)
        y0 = y0.at[1].set(1.0)

        # pycutest sets Q matrix with a specific pattern:
        # Row i has non-zeros at columns (i+1) mod 50 and (2*i+3) mod 50
        # with special handling for certain rows
        q_start = self.N
        for i in range(self.N):
            if i == 24:
                # Row 24 only has Q[24,25] = 1.0
                idx = q_start + i * self.N + 25
                y0 = y0.at[idx].set(1.0)
            elif i == 49:
                # Row 49 only has Q[49,49] = 1.0
                idx = q_start + i * self.N + 49
                y0 = y0.at[idx].set(1.0)
            elif i >= 25 and i < 49:
                # Rows 25-48 have a wrapped pattern
                # First column: 2*(i-25)+1
                col1 = 2 * (i - 25) + 1
                idx = q_start + i * self.N + col1
                y0 = y0.at[idx].set(1.0)
                # Second column: i+1
                col2 = i + 1
                idx = q_start + i * self.N + col2
                y0 = y0.at[idx].set(1.0)
            else:
                # Rows 0-23 follow the original pattern
                # Q[i, i+1] = 1.0
                if i + 1 < self.N:
                    idx = q_start + i * self.N + (i + 1)
                    y0 = y0.at[idx].set(1.0)

                # Q[i, 2*i+3] = 1.0
                if 2 * i + 3 < self.N:
                    idx = q_start + i * self.N + (2 * i + 3)
                    y0 = y0.at[idx].set(1.0)

        return inexact_asarray(y0)

    @property
    def args(self):
        """No additional arguments."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution: eigenvalues 1, ..., N and identity matrix."""
        # For diagonal matrix A with eigenvalues 1, ..., N,
        # the solution is D = diag(1, ..., N) and Q = I
        d_expected = inexact_asarray(jnp.arange(1, self.N + 1))
        q_expected = inexact_asarray(jnp.eye(self.N).flatten())
        return jnp.concatenate([d_expected, q_expected])

    @property
    def expected_objective_value(self):
        """Expected optimal objective value (0 for constrained formulation)."""
        return inexact_asarray(jnp.array(0.0))

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[jnp.ndarray, jnp.ndarray] | None:
        """Nonnegative bounds on all variables as per SIF comment."""
        # The SIF file has comment "nonnegative eigenvalues" which pycutest
        # interprets as lower bounds of 0.0 on ALL variables
        lower = jnp.zeros(self.n)
        upper = jnp.full(self.n, jnp.inf)
        return lower, upper



import jax.numpy as jnp

from ..._problem import AbstractConstrainedQuadraticProblem


class HATFLDH(AbstractConstrainedQuadraticProblem):
    """A test problem from the OPTIMA user manual.

    This is a nonlinear objective with linear constraints.

    Source:
    "The OPTIMA user manual (issue No.8, p. 91)",
    Numerical Optimization Centre, Hatfield Polytechnic (UK), 1989.

    SIF input: Ph. Toint, May 1990.

    classification QLR2-AN-4-7
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def n(self):
        """Number of variables."""
        return 4

    @property
    def y0(self):
        """Initial guess."""
        return jnp.array([1.0, 5.0, 5.0, 1.0])

    @property
    def args(self):
        return None

    def objective(self, y, args):
        """Quadratic objective function: -x1*x3 - x2*x4."""
        del args
        x1, x2, x3, x4 = y[0], y[1], y[2], y[3]

        # Two product terms
        return -x1 * x3 - x2 * x4

    @property
    def bounds(self):
        """Variable bounds: 0 <= x_i <= 5."""
        lower = jnp.zeros(4)
        upper = jnp.full(4, 5.0)
        return lower, upper

    def constraint(self, y):
        """Linear constraints.

        From the SIF file, we have:
        C1-C6: Range constraints (treated as single constraints by pycutest)
        C7: One-sided inequality constraint x1 + x2 + x3 + x4 >= 5.0

        For range constraints, pycutest returns them shifted by lower bound.
        """
        x1, x2, x3, x4 = y[0], y[1], y[2], y[3]

        # No equality constraints
        eq_constraint = None

        # All constraints are inequalities
        inequalities = jnp.array(
            [
                (x1 + x2) - 2.5,  # C1: 2.5 <= x1 + x2 <= 7.5
                (x1 + x3) - 2.5,  # C2: 2.5 <= x1 + x3 <= 7.5
                (x1 + x4) - 2.5,  # C3: 2.5 <= x1 + x4 <= 7.5
                (x2 + x3) - 2.0,  # C4: 2.0 <= x2 + x3 <= 7.0
                (x2 + x4) - 2.0,  # C5: 2.0 <= x2 + x4 <= 7.0
                (x3 + x4) - 1.5,  # C6: 1.5 <= x3 + x4 <= 6.5
                (x1 + x2 + x3 + x4) - 5.0,  # C7: x1 + x2 + x3 + x4 >= 5.0
            ]
        )

        return eq_constraint, inequalities

    @property
    def expected_result(self):
        """Expected result not provided in SIF file."""
        return None

    @property
    def expected_objective_value(self):
        """Expected objective value from SIF file."""
        return jnp.array(-24.4999998)



# TODO: Human review needed
# Attempts made: Multiple formulation attempts to match pycutest
# Suspected issues: Variable ordering or constraint formulation discrepancy
# Resources needed: Deep analysis of pycutest's READING6 implementation
# Note: Passes 23/27 tests. Failures at ones vector for constraints/Jacobians
# The constraint values differ by ~2.2 units consistently at non-zero points

import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class READING6(AbstractConstrainedMinimisation):
    """A nonlinear optimal control problem from Nancy Nichols and Mandy Crossley.
    This problem arises in tide modelling. Course discretization.

    Source: a variant upon a problem in
    S. Lyle and N.K. Nichols,
    "Numerical Methods for Optimal Control Problems with State Constraints",
    Numerical Analysis Report 8/91, Dept of Mathematics,
    University of Reading, UK.

    SIF input: Nick Gould, Feb 1995
    Classification: OOR2-MN-102-50
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})
    n: int = 50

    # Constants
    c: float = 2.41e3
    k: float = 0.95e-3

    # Data arrays G(0:50)
    g_values = jnp.array(
        [
            -2.0627e-03,
            -9.9414e-04,
            -5.7942e-04,
            -3.6153e-04,
            -2.1980e-04,
            -1.1452e-04,
            -4.4283e-05,
            4.0651e-05,
            1.0438e-04,
            1.1511e-04,
            1.4736e-04,
            1.6264e-04,
            1.9018e-04,
            1.9992e-04,
            2.1341e-04,
            2.3507e-04,
            2.4552e-04,
            2.5497e-04,
            2.6397e-04,
            2.7167e-04,
            2.6796e-04,
            2.6980e-04,
            2.6349e-04,
            2.5023e-04,
            2.3040e-04,
            2.0245e-04,
            1.7031e-04,
            1.2704e-04,
            7.9276e-05,
            2.8594e-05,
            -2.3288e-05,
            -7.2000e-05,
            -1.1627e-04,
            -1.5002e-04,
            -1.7517e-04,
            -1.8623e-04,
            -1.8593e-04,
            -1.7524e-04,
            -1.5399e-04,
            -1.2037e-04,
            -7.5545e-05,
            -2.5328e-05,
            2.8266e-05,
            7.9945e-05,
            1.2594e-04,
            1.6135e-04,
            1.8176e-04,
            1.8627e-04,
            1.7264e-04,
            1.4206e-04,
            9.6944e-05,
        ]
    )

    # Data arrays Q(0:50)
    q_values = jnp.array(
        [
            -1.1134e01,
            -1.1046e01,
            -1.0784e01,
            -1.0352e01,
            -9.7564e00,
            -9.0073e00,
            -8.1160e00,
            -7.0968e00,
            -5.9657e00,
            -4.7404e00,
            -3.4405e00,
            -2.0862e00,
            -6.9907e-01,
            6.9910e-01,
            2.0862e00,
            3.4405e00,
            4.7405e00,
            5.9657e00,
            7.0968e00,
            8.1160e00,
            9.0073e00,
            9.7564e00,
            1.0352e01,
            1.0784e01,
            1.1046e01,
            1.1134e01,
            1.1046e01,
            1.0784e01,
            1.0352e01,
            9.7564e00,
            9.0073e00,
            8.1160e00,
            7.0968e00,
            5.9657e00,
            4.7404e00,
            3.4405e00,
            2.0862e00,
            6.9907e-01,
            -6.9910e-01,
            -2.0862e00,
            -3.4405e00,
            -4.7405e00,
            -5.9657e00,
            -7.0968e00,
            -8.1160e00,
            -9.0073e00,
            -9.7564e00,
            -1.0352e01,
            -1.0784e01,
            -1.1046e01,
            -1.1134e01,
        ]
    )

    def objective(self, y, args):
        # Variables are P(0:N) followed by A(0:N)
        p = y[: self.n + 1]
        a = y[self.n + 1 :]

        # Objective: sum of O(j) elements for j=1 to N
        # O(j) uses AP2 element: -A(j) * P(j)^2
        # Coefficients: 1.0 for j=1 to N-1, 0.5 for j=N

        # AP2 elements: -a * p^2
        ap2_elements = -a * p * p

        # Sum with appropriate coefficients
        obj = jnp.sum(ap2_elements[1 : self.n]) + 0.5 * ap2_elements[self.n]

        return obj

    @property
    def y0(self):
        # Initial guess: zeros for P and A
        # Note: P(0) is fixed at 0.0
        return jnp.zeros(2 * (self.n + 1))

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        return self.y0

    @property
    def expected_objective_value(self):
        return None

    @property
    def bounds(self):
        # P(0) is fixed at 0.0
        # P(1:N) are free (unbounded)
        # A(0:N) are bounded in [0, 1]

        lower = jnp.full(2 * (self.n + 1), -jnp.inf)
        upper = jnp.full(2 * (self.n + 1), jnp.inf)

        # P(0) fixed at 0.0
        lower = lower.at[0].set(0.0)
        upper = upper.at[0].set(0.0)

        # A(i) bounds: [0, 1] for i=0 to N
        for i in range(self.n + 1):
            idx = self.n + 1 + i
            lower = lower.at[idx].set(0.0)
            upper = upper.at[idx].set(1.0)

        return lower, upper

    def constraint(self, y):
        # Variables
        p = y[: self.n + 1]
        a = y[self.n + 1 :]

        # Mesh parameters
        delta_t = 10.0 / self.n
        d_half = delta_t * 0.5
        d2_4 = d_half * d_half

        # Derived constants
        # cd_half = self.c * d_half  # Not currently used
        cd2_4 = self.c * d2_4
        cd2_2 = cd2_4 * 2.0
        ck = self.c * self.k
        ckd_half = ck * d_half

        # Special constants
        g0_2g1 = self.g_values[0] + 2.0 * self.g_values[1]
        ggcd2_4 = cd2_4 * g0_2g1
        g0cd2_4 = cd2_4 * self.g_values[0]
        l1 = -1.0 - g0cd2_4
        l2 = -1.0 + ggcd2_4

        # Build equality constraints C(1) to C(N)
        # Note: Using a fully vectorized approach is complex due to triangular summation
        # So we use a hybrid approach with partial vectorization

        equalities = jnp.zeros(self.n)

        for i in range(self.n):
            # C(i+1) constraint
            # Linear terms: L1 * P(i+1) + L2 * P(i)
            constraint_val = l1 * p[i + 1] + l2 * p[i]

            # Summation terms for j=1 to i-1
            if i > 0:
                j_indices = jnp.arange(1, i)
                i_minus_j = i - j_indices
                i_plus_1_minus_j = i_minus_j + 1
                sum_g = self.g_values[i_minus_j] + self.g_values[i_plus_1_minus_j]
                l3 = -1.0 * sum_g * cd2_2
                constraint_val += jnp.sum(l3 * p[j_indices])

            # AP element terms: CKD/2 * (A(i+1)*P(i+1) + A(i)*P(i))
            constraint_val += ckd_half * (a[i + 1] * p[i + 1] + a[i] * p[i])

            # RHS constant: L4 = D/2 * (Q(i) + Q(i+1))
            l4 = d_half * (self.q_values[i] + self.q_values[i + 1])
            constraint_val -= l4

            equalities = equalities.at[i].set(constraint_val)

        return equalities, None



import jax.numpy as jnp

from ..._problem import AbstractNonlinearEquations


class HYPCIR(AbstractNonlinearEquations):
    """Intersection of a circle and an hyperbola.

    Source: problem 214 (p. 68) in
    A.R. Buckley,
    "Test functions for unconstrained minimization",
    TR 1989CS-3, Mathematics, statistics and computing centre,
    Dalhousie University, Halifax (CDN), 1989.

    SIF input: Ph. Toint, Dec 1989.

    classification NQR2-AN-2-2
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def y0(self):
        """Starting point."""
        return jnp.array([0.0, 1.0])

    @property
    def args(self):
        return None

    @property
    def m(self):
        """Number of equations."""
        return 2

    @property
    def n(self):
        """Number of variables."""
        return 2

    def residual(self, y, args):
        """Compute the residuals (vectorized).

        The equations are:
        1. x1 * x2 = 1  (hyperbola)
        2. x1^2 + x2^2 = 4  (circle)

        So the residuals are:
        1. x1 * x2 - 1
        2. x1^2 + x2^2 - 4
        """
        x1, x2 = y[0], y[1]

        r1 = x1 * x2 - 1.0
        r2 = x1**2 + x2**2 - 4.0

        return jnp.array([r1, r2])

    @property
    def expected_result(self):
        """Expected solution (not provided in SIF)."""
        return None

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[jnp.ndarray, jnp.ndarray] | None:
        """No bounds for this problem."""
        return None



import jax.numpy as jnp

from ..._misc import inexact_asarray
from ..._problem import AbstractUnconstrainedMinimisation


class PENALTY1(AbstractUnconstrainedMinimisation):
    """
    PENALTY1 problem.

    This problem is a sum of n+1 least-squares groups, the first n of
    which have only a linear element.
    It Hessian matrix is dense.

    Source:  Problem 23 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    See also Buckley #181 (p. 79).

    SIF input: Ph. Toint, Dec 1989.

    classification SUR2-AN-V-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})
    n: int = 1000

    def objective(self, y, args):
        del args

        # First n groups: 100000 * (x[i] - 1)^2 for i=0 to n-1
        # Each group G(i) has x[i] with coefficient 1.0, constant -1.0, and scale 100000
        # Use float64 to ensure precision
        y = jnp.asarray(y, dtype=jnp.float64)
        linear_terms = y - 1.0
        scaled_squared = jnp.float64(100000.0) * (linear_terms * linear_terms)
        sum1 = jnp.sum(scaled_squared, dtype=jnp.float64)

        # Last group G(M=n+1): (sum of x[i]^2 - 0.25)^2
        # This group has elements E(i) which are x[i]^2, and constant -0.25
        x_squared = y * y
        elements_sum = jnp.sum(x_squared, dtype=jnp.float64)
        group_val = elements_sum - jnp.float64(0.25)
        sum2 = group_val * group_val

        return sum1 + sum2

    @property
    def y0(self):
        # Starting point: X(I) = I for I=1 to N
        return inexact_asarray(jnp.arange(1.0, self.n + 1.0))

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # Solution not provided in detail in SIF file
        return None

    @property
    def expected_objective_value(self):
        # Solution values provided for some values of n
        n = self.n
        if n == 4:
            return jnp.array(2.24997e-4)
        elif n == 10:
            return jnp.array(7.08765e-5)
        else:
            # For other values, solution not provided
            return None

    def num_variables(self):
        return self.n



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: human review required
class HATFLDD(AbstractUnconstrainedMinimisation):
    """An exponential fitting test problem from the OPTIMA user manual.

    Source:
    "The OPTIMA user manual (issue No.8, p. 35)",
    Numerical Optimization Centre, Hatfield Polytechnic (UK), 1989.

    SIF input: Ph. Toint, May 1990.

    Classification: SUR2-AN-3-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y, args):
        del args
        x1, x2, x3 = y

        # Time points
        t_values = jnp.array([0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.75, 0.8, 0.85, 0.9])

        # Data values
        z_values = jnp.array(
            [1.751, 1.561, 1.391, 1.239, 1.103, 0.981, 0.925, 0.8721, 0.8221, 0.7748]
        )

        # Compute model values for each data point
        # The model is exp(t * x3) - x1 * exp(t * x2) + z
        # So the residual is: exp(t * x3) - x1 * exp(t * x2) + z
        residuals = jnp.exp(t_values * x3) - x1 * jnp.exp(t_values * x2) + z_values

        # Objective function is sum of squared residuals
        return jnp.sum(residuals**2)

    @property
    def y0(self):
        # Initial point from SIF file (lines 75-77)
        return jnp.array([1.0, -1.0, 0.0])

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # Not provided in the SIF file
        return None

    @property
    def expected_objective_value(self):
        # According to SIF file (line 120), the optimal objective value is 6.615114D-08
        return jnp.array(6.615114e-08)



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: human review required
class DENSCHNA(AbstractUnconstrainedMinimisation):
    """Dennis-Schnabel problem A.

    This is a 2-dimensional unconstrained optimization problem with
    nonlinear terms including exponentials.

    Source: Problem from "Numerical Methods for Unconstrained Optimization
    and Nonlinear Equations" by J.E. Dennis and R.B. Schnabel, 1983.

    Classification: OUR2-AN-2-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 2  # Number of variables

    def objective(self, y, args):
        del args
        x1, x2 = y

        # From AMPL model: x[1]^4 + (x[1]+x[2])^2 + (-1.0+exp(x[2]))^2
        # Compute powers and exponential once
        x1_sq = x1 * x1
        x1_4 = x1_sq * x1_sq
        exp_x2 = jnp.exp(x2)
        x1_plus_x2 = x1 + x2

        term1 = x1_4
        term2 = x1_plus_x2 * x1_plus_x2
        term3 = (-1.0 + exp_x2) ** 2

        return term1 + term2 + term3

    @property
    def y0(self):
        # Initial values based on problem specification
        return jnp.array([1.0, 1.0])

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # The minimum is at x = [0.0, 0.0]
        return jnp.array([0.0, 0.0])

    @property
    def expected_objective_value(self):
        # At x = [0.0, 0.0]: 0^4 + (0+0)^2 + (-1+exp(0))^2 = 0 + 0 + 0 = 0
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class PALMER5C(AbstractUnconstrainedMinimisation):
    """A linear least squares problem arising from chemical kinetics.

    model: H-N=C=Se TZVP + MP2
    fitting Y to A0 T_0 + A2 T_2 + A4 T_4 + A6 T_6 + A8 T_8 +
                 A10 T_10
    where T_i is the i-th (shifted) Chebyshev polynomial

    Source:
    M. Palmer, Edinburgh, private communication.

    SIF input: Nick Gould, 1992.

    classification: QUR2-RN-6-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 6  # 6 coefficients
    m: int = 12  # 12 data points (X12 through X23)

    @property
    def y0(self):
        # All coefficients start at 1.0
        return jnp.ones(self.n)

    @property
    def args(self):
        # X data values (radians) - only X12 through X23 are used
        x_data = jnp.array(
            [
                0.000000,
                1.570796,
                1.396263,
                1.308997,
                1.221730,
                1.125835,
                1.047198,
                0.872665,
                0.698132,
                0.523599,
                0.349066,
                0.174533,
            ]
        )

        # Y data values (KJmol-1)
        y_data = jnp.array(
            [
                83.57418,
                81.007654,
                18.983286,
                8.051067,
                2.044762,
                0.000000,
                1.170451,
                10.479881,
                25.785001,
                44.126844,
                62.822177,
                77.719674,
            ]
        )

        return (x_data, y_data)

    def objective(self, y, args):
        """Compute the sum of squared residuals using Chebyshev polynomials."""
        a0, a2, a4, a6, a8, a10 = y
        x_data, y_data = args

        # Transform X to Y = 2*X/(B-A) - (A+B)/(B-A) where A = -1.570796, B = 1.570796
        # This maps the interval [A, B] to [-1, 1] for Chebyshev polynomials
        a_val = -1.570796
        b_val = 1.570796
        diff = b_val - a_val
        y_transformed = 2.0 * x_data / diff - (a_val + b_val) / diff

        # Compute Chebyshev polynomials recursively
        # T0 = 1, T1 = y, T_n = 2*y*T_{n-1} - T_{n-2}
        t0 = jnp.ones_like(y_transformed)
        t1 = y_transformed

        # T2 = 2*y*T1 - T0
        t2 = 2.0 * y_transformed * t1 - t0

        # T4 = 2*y*T3 - T2, but we need T3 first
        t3 = 2.0 * y_transformed * t2 - t1
        t4 = 2.0 * y_transformed * t3 - t2

        # T6 = 2*y*T5 - T4, but we need T5 first
        t5 = 2.0 * y_transformed * t4 - t3
        t6 = 2.0 * y_transformed * t5 - t4

        # T8 = 2*y*T7 - T6, but we need T7 first
        t7 = 2.0 * y_transformed * t6 - t5
        t8 = 2.0 * y_transformed * t7 - t6

        # T10 = 2*y*T9 - T8, but we need T9 first
        t9 = 2.0 * y_transformed * t8 - t7
        t10 = 2.0 * y_transformed * t9 - t8

        # Model prediction using even Chebyshev polynomials only
        predicted = a0 * t0 + a2 * t2 + a4 * t4 + a6 * t6 + a8 * t8 + a10 * t10

        # Compute sum of squared residuals
        residuals = predicted - y_data
        return jnp.sum(residuals**2)

    @property
    def expected_result(self):
        # The optimal solution is not explicitly given in the SIF file
        return None

    @property
    def expected_objective_value(self):
        # From the SIF file comment
        return jnp.array(5.0310687e-02)



import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractConstrainedMinimisation


class KSIP(AbstractConstrainedMinimisation):
    """Source: problem from Kortanek and No
    The problem is really a semi-infinite QP
    to appear in SIAM J. Optimization.

    The structure is the following:
    min  "Sum"{ Xj^2/(2j) + Xj/j  ;  j=1,...,n }   subject to
    "Sum"{ t^(j-1)*Xj } ; j=1,...,n  >=  b(t) for all t in [0 1].

    Four examples are considered for n = 20, corresponding to the RHS
    function, b(t) : sin(t), 1/(2-t), exp(t), and tan(t).

    The interval [0 1] is dicretized via steps of 1/1000

    SIF input: A.R. Conn, May 1993

    classification QLR2-AN-20-1001
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Problem parameters
    N = 20  # Number of variables
    M = 1000  # Number of discretization points

    def objective(self, y: Array, args) -> Array:
        """Quadratic objective: sum of Xj^2/(2j) + Xj/j for j=1,...,N"""
        j_values = jnp.arange(1, self.N + 1)
        terms = y**2 / (2.0 * j_values) + y / j_values
        return jnp.sum(terms)

    def constraint(self, y: Array):
        """Semi-infinite constraints discretized at 1000 points.

        For each t in [0, 1] (discretized), we have:
        sum_{j=1}^N t^(j-1) * X_j >= sin(t)

        Converted to inequality constraints <= 0:
        sin(t) - sum_{j=1}^N t^(j-1) * X_j <= 0
        """
        # Generate discretization points: t = i/M for i = 0, 1, ..., M
        t_values = jnp.arange(self.M + 1) / self.M

        constraints = []
        for t in t_values:
            # Compute sum_{j=1}^N t^(j-1) * X_j
            constraint_sum = 0.0
            for j in range(1, self.N + 1):
                xj = y[j - 1]
                constraint_sum += (t ** (j - 1)) * xj

            # sin(t) - sum_{j=1}^N t^(j-1) * X_j <= 0
            constraint_val = jnp.sin(t) - constraint_sum
            constraints.append(constraint_val)

        # No equality constraints
        equality_constraints = jnp.array([])

        # All constraints are inequalities
        inequality_constraints = jnp.array(constraints)

        return equality_constraints, inequality_constraints

    @property
    def y0(self):
        """Initial guess for the optimization problem."""
        # From SIF: all variables start at 2.0
        return jnp.full(self.N, 2.0)

    @property
    def bounds(self):
        """Variable bounds. All variables are free."""
        # From SIF: FR KSIP 'DEFAULT' means all variables are free
        return None

    @property
    def args(self):
        """Additional arguments for the objective function."""
        return None

    @property
    def expected_result(self):
        """Expected result of the optimization problem."""
        # This is a complex semi-infinite programming problem
        # The exact solution is not straightforward to compute
        return jnp.zeros(self.N)

    @property
    def expected_objective_value(self):
        """Expected value of the objective at the solution."""
        # From SIF comments, multiple solutions are mentioned
        # Using the first one as a reference
        return jnp.array(0.57579024357147)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class READING1(AbstractConstrainedMinimisation):
    """A nonlinear optimal control problem from Nancy Nichols with a given initial
    condition. This problem arises in tide modelling.

    Source:
    S. Lyle and N.K. Nichols,
    "Numerical Methods for Optimal Control Problems with State Constraints",
    Numerical Analysis Report 8/91, Dept of Mathematics,
    University of Reading, UK.

    SIF input: Nick Gould, July 1991.
    Classification: OOR2-MN-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})
    n: int = 2000  # Number of discretized points
    pi: float = 3.141592653589
    a: float = 0.07716
    h: float = 1.0 / 2000  # 1.0 / n
    two_pi: float = 2.0 * 3.141592653589  # 2.0 * pi
    half_a_inv: float = 0.5 / 0.07716  # 0.5 / a
    h_inv: float = 1.0 / (1.0 / 2000)  # 1.0 / h
    h_half: float = (1.0 / 2000) / 2.0  # h / 2.0
    x0_fixed: float = 0.25  # Fixed value for x[0]

    def objective(self, y, args):
        # Variables are interleaved: x[0], u[0], x[1], u[1], ..., x[n], u[n]
        # Extract x and u from interleaved format
        x = y[::2]  # Every even index: x[0], x[1], ..., x[n]
        u = y[1::2]  # Every odd index: u[0], u[1], ..., u[n]

        # Vectorized computation of ENERGY elements I(0) to I(N)
        i_indices = jnp.arange(self.n + 1, dtype=y.dtype)
        t_values = i_indices * self.h
        cos_2pi_t = jnp.cos(self.two_pi * t_values)

        # ENERGY element: HOVER2 * U * (X - C2PIT)^2
        energy_elements = self.h_half * u * (x - cos_2pi_t) ** 2

        # GROUP USES: Each group I(i) for i=1 to N gets -I(i) - I(i-1)
        # This is equivalent to -2 * sum(I(1:N-1)) - I(0) - I(N)
        obj = (
            -2.0 * jnp.sum(energy_elements[1 : self.n])
            - energy_elements[0]
            - energy_elements[self.n]
        )

        return obj

    @property
    def y0(self):
        # Initial guess: all zeros (SIF file defaults)
        # Variables are interleaved: x[0], u[0], x[1], u[1], ..., x[n], u[n]
        # Total: 2*(n+1) = 4002 variables
        return jnp.zeros(2 * (self.n + 1))

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # The expected result would need to be determined from running the problem
        # For now, returning the initial guess
        return self.y0

    @property
    def expected_objective_value(self):
        # Expected objective value needs to be determined
        return None

    @property
    def bounds(self):
        # Bounds are interleaved: x[0], u[0], x[1], u[1], ..., x[n], u[n]
        lower = jnp.zeros(2 * (self.n + 1))
        upper = jnp.zeros(2 * (self.n + 1))

        # Set bounds for interleaved variables
        for i in range(self.n + 1):
            # x[i] bounds
            if i == 0:
                lower = lower.at[2 * i].set(self.x0_fixed)  # x[0] fixed
                upper = upper.at[2 * i].set(self.x0_fixed)
            else:
                lower = lower.at[2 * i].set(-0.5)
                upper = upper.at[2 * i].set(0.5)

            # u[i] bounds
            lower = lower.at[2 * i + 1].set(0.0)
            upper = upper.at[2 * i + 1].set(1.0)

        return lower, upper

    def constraint(self, y):
        # Variables are interleaved: x[0], u[0], x[1], u[1], ..., x[n], u[n]
        x = y[::2]  # Every even index: x[0], x[1], ..., x[n]
        u = y[1::2]  # Every odd index: u[0], u[1], ..., u[n]

        # Vectorized constraint computation for i=1 to n
        i_indices = jnp.arange(1, self.n + 1, dtype=y.dtype)
        t_i = i_indices * self.h
        t_im1 = (i_indices - 1) * self.h
        cos_2pi_ti = jnp.cos(self.two_pi * t_i)
        cos_2pi_tim1 = jnp.cos(self.two_pi * t_im1)

        # Linear part: (X(I) - X(I-1))/H
        constraints = (x[1:] - x[:-1]) * self.h_inv

        # CCTI * U(I) + CCTI-1 * U(I-1) where CCTI = cos(2πt[i]) * (-1/(2a))
        constraints += cos_2pi_ti * (-self.half_a_inv) * u[1:]
        constraints += cos_2pi_tim1 * (-self.half_a_inv) * u[:-1]

        # Nonlinear part: NC(I) + NC(I-1) where NC = P * U * X with P = 1/(2a)
        constraints += self.half_a_inv * u[1:] * x[1:]
        constraints += self.half_a_inv * u[:-1] * x[:-1]

        return constraints, None



import jax.numpy as jnp

from ..._misc import inexact_asarray
from ..._problem import AbstractConstrainedMinimisation


class SIPOW1(AbstractConstrainedMinimisation):
    """
    SIPOW1 problem.

    This is a discretization of a semi-infinite programming problem, of
    minimizing the variable x_2 within a circle of radius 1. The circle
    is replaced by a discrete set of equally-spaced supporting tangents.

    Source: problem 1 in
    M. J. D. Powell,
    "Log barrier methods for semi-infinite programming calculations"
    Numerical Analysis Report DAMTP 1992/NA11, U. of Cambridge, UK.

    SIF input: A. R. Conn and Nick Gould, August 1993

    classification LLR2-AN-2-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})
    m: int = 2000

    def __check_init__(self):
        # No specific constraints on m for SIPOW1
        pass

    def objective(self, y, args):
        del args
        x1, x2 = y
        # Linear objective: minimize x2
        return x2

    @property
    def y0(self):
        # Starting point
        return inexact_asarray(jnp.array([0.8, 0.5]))

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # Solution not provided in detail in SIF file
        return None

    @property
    def expected_objective_value(self):
        # Solution value provided as -1.0
        return jnp.array(-1.0)

    def num_variables(self):
        return 2

    @property
    def bounds(self):
        # All variables are free (unbounded)
        return None

    def constraint(self, y):
        x1, x2 = y
        m = self.m

        # Generate angles for the constraints
        j_vals = inexact_asarray(jnp.arange(1, m + 1))
        angles = 2.0 * jnp.pi * j_vals / inexact_asarray(m)
        cos_vals = jnp.cos(angles)
        sin_vals = jnp.sin(angles)

        # Inequality constraints (in >= 0 form):
        # C(j): x1 * cos(2*pi*j/m) + x2 * sin(2*pi*j/m) >= -1
        # In pycutest form: x1 * cos(2*pi*j/m) + x2 * sin(2*pi*j/m) + 1
        ineq_constraint = x1 * cos_vals + x2 * sin_vals + 1.0

        # Equality constraints: none
        eq_constraint = None

        return eq_constraint, ineq_constraint



import jax.numpy as jnp

from ..._problem import AbstractNonlinearEquations


class CLUSTER(AbstractNonlinearEquations):
    """CLUSTER problem as a nonlinear equations formulation.

    Source: problem 207 in
    A.R. Buckley,
    "Test functions for unconstrained minimization",
    TR 1989CS-3, Mathematics, statistics and computing centre,
    Dalhousie University, Halifax (CDN), 1989.

    SIF input: Ph. Toint, Dec 1989.

    Classification: NOR2-AN-2-2
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def n(self):
        """Number of variables."""
        return 2

    def num_residuals(self):
        """Number of residual equations."""
        return 2  # Two residual equations

    def residual(self, y, args):
        """Compute the residuals.

        The residuals are:
        G1 = (X1 - X2^2) * (X1 - sin(X2))
        G2 = (cos(X2) - X1) * (X2 - cos(X1))
        """
        del args
        x1, x2 = y

        # Element A: G1
        f1_a = x1 - x2**2
        f2_a = x1 - jnp.sin(x2)
        g1 = f1_a * f2_a

        # Element B: G2
        f1_b = jnp.cos(x2) - x1
        f2_b = x2 - jnp.cos(x1)
        g2 = f1_b * f2_b

        return jnp.array([g1, g2])

    @property
    def y0(self):
        """Initial guess."""
        return jnp.zeros(2)

    @property
    def args(self):
        """No additional arguments."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution."""
        # Not provided in SIF file
        return None

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[jnp.ndarray, jnp.ndarray] | None:
        """No bounds for this problem."""
        return None



import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractNonlinearEquations


class LUKSAN15(AbstractNonlinearEquations):
    """Problem 15 (sparse signomial) from Luksan.

    This is a system of nonlinear equations from the paper:
    L. Luksan
    "Hybrid methods in large sparse nonlinear least squares"
    J. Optimization Theory & Applications 89(3) 575-595 (1996)

    SIF input: Nick Gould, June 2017.

    classification NOR2-AN-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})
    s: int = 49  # Seed for dimensions (default from SIF)

    @property
    def n(self) -> int:
        """Number of variables: 2*S + 2."""
        return 2 * self.s + 2

    @property
    def m(self) -> int:
        """Number of equations: 4*S."""
        return 4 * self.s

    @property
    def y0(self) -> Array:
        """Initial guess: pattern (-0.8, 1.2, -1.2, 0.8) repeated."""
        pattern = jnp.array([-0.8, 1.2, -1.2, 0.8], dtype=jnp.float64)
        # Repeat pattern to cover all n variables
        full_pattern = jnp.tile(pattern, (self.n + 3) // 4)[: self.n]
        return full_pattern

    @property
    def args(self):
        """No additional arguments."""
        return None

    def residual(self, y: Array, args) -> Array:
        """Compute the residual vector."""
        del args  # Not used

        x = y
        s = self.s

        # Data values
        Y = jnp.array([35.8, 11.2, 6.2, 4.4], dtype=jnp.float64)

        # Create indices for vectorized computation
        # For each block j in range(s), variables are at i = 2*j, so:
        # x1 = x[2*j], x2 = x[2*j+1], x3 = x[2*j+2], x4 = x[2*j+3]
        j_indices = jnp.arange(s)
        i_indices = 2 * j_indices

        # Extract variables for all blocks
        x1 = x[i_indices]  # x[i] for all blocks
        x2 = x[i_indices + 1]  # x[i+1] for all blocks
        x3 = x[i_indices + 2]  # x[i+2] for all blocks
        x4 = x[i_indices + 3]  # x[i+3] for all blocks

        # Compute signomial term for all blocks: x1 * x2^2 * x3^3 * x4^4
        signom_vals = x1 * (x2**2) * (x3**3) * (x4**4)  # shape: (s,)

        # Create arrays for p and l values
        p_vals = jnp.array([1, 2, 3], dtype=jnp.float64)  # p = 1, 2, 3
        l_vals = jnp.array([1, 2, 3, 4], dtype=jnp.float64)  # l = 1, 2, 3, 4

        # Create meshgrids for vectorized computation
        P, L, J = jnp.meshgrid(
            p_vals, l_vals, j_indices, indexing="ij"
        )  # shapes: (3, 4, s)

        # Compute p2ol and pli for all combinations
        P2OL = P**2 / L  # shape: (3, 4, s)
        PLI = 1.0 / (P * L)  # shape: (3, 4, s)

        # Expand signom_vals to match the shape
        signom_expanded = signom_vals[None, None, :]  # shape: (1, 1, s)

        # Apply sign based on whether signom_val is positive
        sign_p = jnp.where(signom_expanded > 0, 1.0, -1.0)
        p_val = signom_expanded * sign_p  # This gives |signom_val|, shape: (1, 1, s)

        # Compute F = p2ol * p_val^pli for all combinations
        F_vals = P2OL * (p_val**PLI)  # shape: (3, 4, s)

        # Sum over p (axis=0) to get equation sums for each (l, j)
        eq_sums = jnp.sum(F_vals, axis=0)  # shape: (4, s)

        # Subtract Y values (broadcast Y to match shape)
        Y_expanded = Y[:, None]  # shape: (4, 1)
        residuals_matrix = eq_sums - Y_expanded  # shape: (4, s)

        # Flatten in the correct order: for each j, then for each l
        # The original order is: j=0,l=1; j=0,l=2; j=0,l=3; j=0,l=4; j=1,l=1; ...
        # residuals_matrix is (l, j), so we need to transpose and flatten
        residuals = residuals_matrix.T.flatten()  # shape: (4*s,)

        return residuals

    @property
    def expected_result(self) -> Array | None:
        """Expected optimal solution."""
        # No specific expected result provided in SIF
        return None

    @property
    def expected_objective_value(self) -> Array | None:
        """Expected objective value (sum of squares)."""
        # For nonlinear equations, the expected value is 0
        return jnp.array(0.0)

    def constraint(self, y: Array):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """No bounds for this problem."""
        return None



import jax.numpy as jnp

from ..._problem import AbstractNonlinearEquations


class POWELLSE(AbstractNonlinearEquations):
    """POWELLSE problem - The extended Powell singular problem.

    This problem is a sum of n/4 sets of four terms, each of which is
    assigned its own group. This is a nonlinear equation version
    of problem POWELLSG.

    Source: Problem 13 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    See also Toint#19, Buckley#34 (p.85)

    SIF input: Ph. Toint, Dec 1989.
    Modification as a set of nonlinear equations: Nick Gould, Oct 2015.

    Classification: NOR2-AN-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def n(self):
        """Number of variables (must be a multiple of 4)."""
        return 4  # Default value

    def num_residuals(self):
        """Number of residuals."""
        # Each set of 4 variables produces 4 residuals
        return self.n

    def residual(self, y, args):
        """Compute the residuals."""
        del args
        x = y

        # Ensure n is a multiple of 4
        assert self.n % 4 == 0, "n must be a multiple of 4"

        # Reshape x to process groups of 4 variables at once
        x_groups = x.reshape(-1, 4)  # Shape: (n/4, 4)

        # Extract components for all groups
        x0 = x_groups[:, 0]  # x(i)
        x1 = x_groups[:, 1]  # x(i+1)
        x2 = x_groups[:, 2]  # x(i+2)
        x3 = x_groups[:, 3]  # x(i+3)

        # Compute residuals for all groups vectorized
        # G(i): x(i) + 10*x(i+1) = 0
        g1 = x0 + 10.0 * x1

        # G(i+1): 5 * (x(i+2) - x(i+3)) = 0
        # Note: SIF has SCALE 0.2, which means multiply by 5 = 1/0.2
        g2 = 5.0 * (x2 - x3)

        # G(i+2): (x(i+1) - 2*x(i+2))^2 = 0
        g3 = (x1 - 2.0 * x2) ** 2

        # G(i+3): 10 * (x(i) - x(i+3))^2 = 0
        # Note: SIF has SCALE 0.1, and squared term gives factor 10
        g4 = 10.0 * (x0 - x3) ** 2

        # Stack and flatten to return residuals in the correct order
        residuals = jnp.stack([g1, g2, g3, g4], axis=1).ravel()

        return residuals

    @property
    def y0(self):
        """Initial guess."""
        # Create the repeating pattern [3.0, -1.0, 0.0, 1.0]
        pattern = jnp.array([3.0, -1.0, 0.0, 1.0])
        # Tile it to match the problem dimension
        x0 = jnp.tile(pattern, self.n // 4)
        return x0

    @property
    def args(self):
        """Additional arguments (none for this problem)."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution (not provided in SIF)."""
        return None

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[jnp.ndarray, jnp.ndarray] | None:
        """No bounds for this problem."""
        return None



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


# TODO: Human review needed - constraint function inconsistencies
# There appear to be minor but relevant inconsistencies in the constraint
# function implementation, likely due to an indexing error. The constraint
# values don't match pycutest exactly, suggesting the indices used in the
# constraint formulation may be off by one or have other systematic errors.
class LUKVLE12(AbstractConstrainedMinimisation):
    """LUKVLE12 - Chained HS47 problem.

    Problem 5.12 from Luksan and Vlcek test problems.

    The objective is a chained HS47 function:
    f(x) = Σ[i=1 to (n-1)/4] [(x_{j+1} - x_{j+2})^2 + (x_{j+2} - x_{j+3})^2 +
                               (x_{j+3} - x_{j+4})^4 + (x_{j+4} - x_{j+5})^4]
    where j = 4(i-1)

    Subject to equality constraints (for K = 1, 4, 7, 10, ...):
    c_K(x) = x_K + x_{K+1}^2 + x_{K+2}^2 - 3 = 0
    c_{K+1}(x) = x_{K+1} + x_{K+3} + x_{K+2}^2 - 1 = 0
    c_{K+2}(x) = x_K * x_{K+4} - 1 = 0
    Special case: c_3(x) = -x_1 * x_5 + 1 = 0 (sign flipped)
    where n_C = 3(n-1)/4

    Starting point:
    x_i = 2.0 for i ≡ 1 (mod 4)
    x_i = 1.5 for i ≡ 2 (mod 4)
    x_i = -1.0 for i ≡ 3 (mod 4)
    x_i = 0.5 for i ≡ 0 (mod 4)

    Source: L. Luksan and J. Vlcek,
    "Sparse and partially separable test problems for
    unconstrained and equality constrained optimization",
    Technical Report 767, Inst. Computer Science, Academy of Sciences
    of the Czech Republic, 182 07 Prague, Czech Republic, 1999

    SIF input: Nick Gould, April 2001

    Classification: OOR2-AY-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 9997  # Default dimension, must satisfy (n-1) divisible by 4

    def objective(self, y, args):
        del args
        n = len(y)
        # Chained HS47 function - vectorized

        # We need groups of 5 consecutive elements starting at indices 0, 4, 8, ...
        num_groups = (n - 1) // 4
        if num_groups == 0:
            return jnp.array(0.0)

        # Create indices for the start of each group
        # Since we know n >= 5 and num_groups = (n-1)//4,
        # all groups should be valid because group i needs elements from 4*i to 4*i+4
        # and the last group starts at 4*(num_groups-1) = 4*((n-1)//4 - 1) < n-4
        i = jnp.arange(num_groups)
        j = 4 * i  # j values in 0-based

        # Extract elements for all groups
        x_j1 = y[j]  # y[j]
        x_j2 = y[j + 1]  # y[j+1]
        x_j3 = y[j + 2]  # y[j+2]
        x_j4 = y[j + 3]  # y[j+3]
        x_j5 = y[j + 4]  # y[j+4]

        # Compute all terms vectorized
        terms = (
            (x_j1 - x_j2) ** 2
            + (x_j2 - x_j3) ** 2
            + (x_j3 - x_j4) ** 4
            + (x_j4 - x_j5) ** 4
        )

        return jnp.sum(terms)

    @property
    def y0(self):
        # Starting point
        y = jnp.zeros(self.n)
        # x_i = 2.0 for i ≡ 1 (mod 4) -> 0-based: i ≡ 0 (mod 4)
        y = y.at[::4].set(2.0)
        # x_i = 1.5 for i ≡ 2 (mod 4) -> 0-based: i ≡ 1 (mod 4)
        y = y.at[1::4].set(1.5)
        # x_i = -1.0 for i ≡ 3 (mod 4) -> 0-based: i ≡ 2 (mod 4)
        y = y.at[2::4].set(-1.0)
        # x_i = 0.5 for i ≡ 0 (mod 4) -> 0-based: i ≡ 3 (mod 4)
        y = y.at[3::4].set(0.5)
        return y

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # Solution pattern based on problem structure
        return None  # Unknown exact solution

    @property
    def expected_objective_value(self):
        return None  # Unknown exact objective value

    @property
    def bounds(self):
        return None

    def constraint(self, y):
        n = len(y)
        n_c = 3 * (n - 1) // 4

        if n_c == 0:
            return jnp.array([]), None

        # Build constraints using K indices from SIF file
        # K takes values 1, 4, 7, 10, ... (in 1-based) with DI K 3
        constraints = []

        # Process each K value
        k = 1
        while len(constraints) < n_c:
            # Convert to 0-based
            k_idx = k - 1

            # Type 1: C(K) = X(K) + X(K+1)² + X(K+2)² - 3 = 0
            if len(constraints) < n_c and k_idx + 2 < n:
                c1 = y[k_idx] + y[k_idx + 1] ** 2 + y[k_idx + 2] ** 2 - 3
                constraints.append(c1)

            # Type 2: C(K+1) = X(K+1) + X(K+3) + X(K+2)² - 1 = 0
            if len(constraints) < n_c and k_idx + 3 < n:
                c2 = y[k_idx + 1] + y[k_idx + 3] + y[k_idx + 2] ** 2 - 1
                constraints.append(c2)

            # Type 3: C(K+2) = X(K) * X(K+4) - 1 = 0
            if len(constraints) < n_c and k_idx + 4 < n:
                c3 = y[k_idx] * y[k_idx + 4] - 1
                # Special case: First type 3 constraint (K=1, C(3)) needs sign flip
                if k == 1:
                    c3 = -c3
                constraints.append(c3)

            # Next K value (K increments by 3)
            k += 3

        # Return constraints as a JAX array
        return jnp.array(constraints), None



"""
A test problem from the OPTIMA user manual.

Source:
"The OPTIMA user manual (issue No.8, p. 26)",
Numerical Optimization Centre, Hatfield Polytechnic (UK), 1989.

SIF input: Ph. Toint, May 1990.

classification SBR2-AN-25-0
"""

import jax.numpy as jnp

from ..._problem import AbstractBoundedMinimisation


class HATFLDC(AbstractBoundedMinimisation):
    @property
    def name(self) -> str:
        return "HATFLDC"

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 25  # 25 variables

    @property
    def y0(self):
        # All variables start at 0.9
        return jnp.full(self.n, 0.9)

    @property
    def args(self):
        return None

    def objective(self, y, args):
        del args

        # The objective function has the structure:
        # G(1) = X(1) - 1.0
        # G(i) = X(i+1) - X(i)^2 for i = 2, ..., N-1
        # G(N) = X(N) - 1.0
        # The objective is sum of G(i)^2 (L2 group type)

        obj = 0.0

        # G(1) = X(1) - 1.0
        g1 = y[0] - 1.0
        obj += g1 * g1

        # G(i) = X(i+1) - X(i)^2 for i = 2, ..., N-1
        for i in range(1, self.n - 1):
            gi = y[i + 1] - y[i] * y[i]
            obj += gi * gi

        # G(N) = X(N) - 1.0
        gn = y[self.n - 1] - 1.0
        obj += gn * gn

        return jnp.array(obj)

    @property
    def bounds(self):
        # Default lower bound is 0.0 in CUTEst when not specified
        # All variables have upper bound 10.0 except X(N) which is free
        lower = jnp.zeros(self.n)
        upper = jnp.full(self.n, 10.0)
        # X(N) is free (XR HATFLDC X(N))
        lower = lower.at[self.n - 1].set(-jnp.inf)
        upper = upper.at[self.n - 1].set(jnp.inf)
        return lower, upper

    @property
    def expected_result(self):
        # The optimal solution is not explicitly given in the SIF file
        return None

    @property
    def expected_objective_value(self):
        # According to the SIF file comment (line 91),
        # the optimal objective value is 0.0
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class TWIRISM1(AbstractConstrainedMinimisation):
    """
    TWIRISM1 - Small nuclear reactor core reload pattern optimization problem.

    # TODO: Human review needed
    # Attempts made: [
    #   1. Fixed initial values from SIF file (hardcoded pycutest array)
    #   2. Implemented basic constraint structure
    #      (Sumi, Sumlm, Norm, Kern, Burn, Plac, Dia)
    #   3. Added complete G matrix (76 non-zero entries) and V vector from SIF
    #   4. Fixed k/phi variable extraction (node-major, interleaved per time step)
    #   5. Corrected bounds (Kfresh=1.2 for this problem)
    #   6. Implemented trilinear elements xa, xb, xc using einsum operations
    # ]
    # Suspected issues: [
    #   - Trilinear ELEMENT variables (xa,xb,xc) formulation not matching
    #     SIF exactly
    #   - xa(i,j,m) = x(i,2,m) * x(j,1,m) * k(j,t_eoc) structure needs
    #     validation
    #   - Constraint values differ from pycutest (Sumi/Sumlm match ~1e-4,
    #     others have larger discrepancies)
    #   - Complex group-separable structure requires deep SIF ELEMENT analysis
    #   - All test categories fail: constraints, constraint Jacobians
    # ]
    # Resources needed: [
    #   - Detailed analysis of SIF ELEMENT and GROUP sections
    #   - Reference implementation or documentation for trilinear nuclear
    #     reactor constraints
    #   - Verification of variable indexing and time step associations
    #   - Comparison with AMPL .mod file if available
    # ]

    A relaxation of a nonlinear integer programming problem for finding
    optimal nuclear reactor core reload patterns (small size version).

    Source:
    A.J. Quist, E. de Klerk, C. Roos, T. Terlaky, R. van Geemert,
    J.E. Hoogenboom, T.Illes,
    "Finding Optimal Nuclear Reactor Core Reload Patterns Using Nonlinear
    Optimization and Search Heuristics",
    Delft University, January 1998.

    classification LOI2-RN-343-313

    SIF input: Arie Quist, Delft, 1998.
    """

    # Problem dimensions
    Nnod: int = 14  # Number of nodes
    Nred: int = 12  # Number of reduced nodes
    Ndia: int = 2  # Number of diagonal sets
    Ntim: int = 6  # Number of time steps
    Nage: int = 4  # Number of age groups
    Ntra: int = 3  # Number of trajectories

    # Required attributes for the abstract class
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Pre-computed constants
    Kfresh: float = 1.2
    alp_Pc_Dt: float = 0.01

    @property
    def initial_guess(self) -> jnp.ndarray:
        """Initial guess matching pycutest structure exactly."""
        # Use the exact pycutest x0 array
        pycutest_x0 = jnp.array(
            [
                0.0333,
                0.0333,
                0.0333,
                0.1849,
                0.1849,
                0.1849,
                0.1151,
                0.1151,
                0.1151,
                0.0,
                0.0,
                0.0,
                0.0333,
                0.0333,
                0.0333,
                0.1849,
                0.1849,
                0.1849,
                0.1151,
                0.1151,
                0.1151,
                0.0,
                0.0,
                0.0,
                0.1,
                0.1,
                0.1,
                0.1349,
                0.1349,
                0.1349,
                0.0651,
                0.0651,
                0.0651,
                0.0333,
                0.0333,
                0.0333,
                0.1,
                0.1,
                0.1,
                0.1349,
                0.1349,
                0.1349,
                0.0651,
                0.0651,
                0.0651,
                0.0333,
                0.0333,
                0.0333,
                0.1,
                0.1,
                0.1,
                0.1349,
                0.1349,
                0.1349,
                0.0651,
                0.0651,
                0.0651,
                0.0333,
                0.0333,
                0.0333,
                0.0,
                0.0,
                0.0,
                0.0083,
                0.0083,
                0.0083,
                0.025,
                0.025,
                0.025,
                0.3,
                0.3,
                0.3,
                0.0333,
                0.0333,
                0.0333,
                0.1849,
                0.1849,
                0.1849,
                0.1151,
                0.1151,
                0.1151,
                0.0,
                0.0,
                0.0,
                0.3,
                0.3,
                0.3,
                0.0083,
                0.0083,
                0.0083,
                0.025,
                0.025,
                0.025,
                0.0,
                0.0,
                0.0,
                0.3,
                0.3,
                0.3,
                0.0083,
                0.0083,
                0.0083,
                0.025,
                0.025,
                0.025,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0333,
                0.0333,
                0.0333,
                0.1186,
                0.1186,
                0.1186,
                0.1814,
                0.1814,
                0.1814,
                0.0333,
                0.0333,
                0.0333,
                0.1256,
                0.1256,
                0.1256,
                0.0558,
                0.0558,
                0.0558,
                0.1186,
                0.1186,
                0.1186,
                0.0,
                0.0,
                0.0,
                0.0333,
                0.0333,
                0.0333,
                0.3,
                0.3,
                0.3,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0083,
                0.0083,
                0.0083,
                0.025,
                0.025,
                0.025,
                0.3,
                0.3,
                0.3,
                0.0333,
                0.0333,
                0.0333,
                0.1256,
                0.1256,
                0.1256,
                0.0558,
                0.0558,
                0.0558,
                0.1186,
                0.1186,
                0.1186,
                1.0231,
                0.0855,
                1.0098,
                0.088,
                0.9962,
                0.0903,
                0.9824,
                0.0922,
                0.9686,
                0.0939,
                0.9547,
                0.0954,
                1.0231,
                0.0997,
                1.0075,
                0.1008,
                0.992,
                0.1016,
                0.9766,
                0.1023,
                0.9613,
                0.1028,
                0.9462,
                0.1031,
                1.0595,
                0.1206,
                1.04,
                0.1207,
                1.0208,
                0.1205,
                1.002,
                0.1203,
                0.9836,
                0.1199,
                0.9655,
                0.1195,
                1.0595,
                0.1008,
                1.0432,
                0.1032,
                1.0267,
                0.1054,
                1.0102,
                0.1076,
                0.9936,
                0.1095,
                0.9769,
                0.1113,
                1.0595,
                0.0534,
                1.0509,
                0.0573,
                1.0417,
                0.0612,
                1.0319,
                0.0653,
                1.0216,
                0.0695,
                1.0108,
                0.0738,
                0.9546,
                0.0147,
                0.9524,
                0.0165,
                0.95,
                0.0184,
                0.9474,
                0.0204,
                0.9444,
                0.0227,
                0.9411,
                0.0251,
                1.0231,
                0.1194,
                1.0045,
                0.1185,
                0.9863,
                0.1176,
                0.9685,
                0.1166,
                0.9513,
                0.1156,
                0.9344,
                0.1147,
                1.1779,
                0.1505,
                1.1508,
                0.1483,
                1.1247,
                0.1461,
                1.0996,
                0.1441,
                1.0753,
                0.1422,
                1.052,
                0.1403,
                1.1779,
                0.1151,
                1.1572,
                0.1166,
                1.1365,
                0.118,
                1.116,
                0.1193,
                1.0957,
                0.1206,
                1.0755,
                0.1218,
                0.9636,
                0.0409,
                0.9576,
                0.0433,
                0.9513,
                0.0458,
                0.9446,
                0.0484,
                0.9376,
                0.0511,
                0.9303,
                0.0539,
                1.0074,
                0.098,
                0.9923,
                0.0982,
                0.9774,
                0.0985,
                0.9627,
                0.0989,
                0.9481,
                0.0993,
                0.9338,
                0.0997,
                0.97,
                0.0559,
                0.9617,
                0.0578,
                0.9533,
                0.0598,
                0.9445,
                0.0619,
                0.9356,
                0.0641,
                0.9264,
                0.0664,
                0.9546,
                0.0207,
                0.9516,
                0.0222,
                0.9483,
                0.0238,
                0.9449,
                0.0255,
                0.9412,
                0.0273,
                0.9373,
                0.0293,
                1.0074,
                0.0245,
                1.0036,
                0.0262,
                0.9996,
                0.028,
                0.9953,
                0.0299,
                0.9908,
                0.032,
                0.9859,
                0.0343,
                1.0558,
                1.0373,
                1.0193,
                1.0017,
                0.9846,
                0.9678,
                0.01,
            ]
        )

        return pycutest_x0

    @property
    def y0(self) -> jnp.ndarray:
        """Initial guess."""
        return self.initial_guess

    @property
    def bounds(self):
        """Variable bounds matching pycutest structure."""
        Nnod = self.Nnod
        Ntim = self.Ntim
        Nage = self.Nage
        Ntra = self.Ntra

        # Sizes
        n_x = Nnod * Nage * Ntra  # 168
        n_kphi = 2 * Nnod * Ntim  # 168 (interleaved k and phi)
        n_keff = Ntim  # 6
        n_eps = 1  # 1

        # Physical constants
        Kfresh = 1.2

        # Bounds for x variables: [0, 1]
        x_lower = jnp.zeros(n_x)
        x_upper = jnp.ones(n_x)

        # Bounds for interleaved k and phi
        # Create interleaved bounds array
        kphi_lower = jnp.zeros(n_kphi)
        kphi_upper = jnp.zeros(n_kphi)

        # Set bounds for k and phi in interleaved pattern
        for i in range(Nnod):
            for t in range(Ntim):
                idx = 2 * (i * Ntim + t)
                # k bounds: [0, Kfresh]
                kphi_lower = kphi_lower.at[idx].set(0.0)
                kphi_upper = kphi_upper.at[idx].set(Kfresh)
                # phi bounds: [0.0001, 1.0]
                kphi_lower = kphi_lower.at[idx + 1].set(0.0001)
                kphi_upper = kphi_upper.at[idx + 1].set(1.0)

        # Bounds for keff: [0, Kfresh]
        keff_lower = jnp.zeros(n_keff)
        keff_upper = jnp.full(n_keff, Kfresh)

        # Bounds for epsilon: [0, 100]
        eps_lower = jnp.zeros(n_eps)
        eps_upper = jnp.full(n_eps, 100.0)

        lower = jnp.concatenate([x_lower, kphi_lower, keff_lower, eps_lower])
        upper = jnp.concatenate([x_upper, kphi_upper, keff_upper, eps_upper])

        return lower, upper

    def objective(self, y: jnp.ndarray, args=None) -> jnp.ndarray:
        """Objective: maximize keff(Ntim-1) - w1*epsilon."""
        Nnod = self.Nnod
        Ntim = self.Ntim
        Nage = self.Nage
        Ntra = self.Ntra

        # Extract variables matching pycutest structure
        n_x = Nnod * Nage * Ntra  # 168
        n_kphi = 2 * Nnod * Ntim  # 168 (interleaved k and phi)

        keff_start = n_x + n_kphi
        keff = y[keff_start : keff_start + Ntim]
        epsilon = y[-1]

        # Weight for epsilon penalty
        w1 = 1.0

        # Maximize keff(Ntim-1) which means minimize -keff(Ntim-1)
        return -keff[Ntim - 1] + w1 * epsilon

    def constraint(self, y: jnp.ndarray) -> tuple[jnp.ndarray, jnp.ndarray]:
        """Constraints for the nuclear reactor problem."""
        Nnod = self.Nnod
        Ntim = self.Ntim
        Nage = self.Nage
        Ntra = self.Ntra

        # Extract variables matching pycutest's interleaved structure
        n_x = Nnod * Nage * Ntra  # 168
        x = y[:n_x].reshape((Nnod, Nage, Ntra))

        # k and phi are interleaved in pycutest
        # Each node has: k[i,0], phi[i,0], k[i,1], phi[i,1], k[i,2], phi[i,2], ...
        n_kphi = 2 * Nnod * Ntim  # 168
        kphi_section = y[n_x : n_x + n_kphi]
        kphi_node_major = kphi_section.reshape((Nnod, 2 * Ntim))
        k = kphi_node_major[:, ::2]  # Every other column starting from 0 (k values)
        phi = kphi_node_major[
            :, 1::2
        ]  # Every other column starting from 1 (phi values)

        keff = y[n_x + n_kphi : n_x + n_kphi + Ntim]
        epsilon = y[-1]

        # Constants
        V = jnp.array(
            [0.5, 1.0, 1.0, 1.0, 1.0, 1.0, 0.5, 1.0, 1.0, 1.0, 0.5, 1.0, 1.0, 0.5]
        )
        G = _build_G_matrix()
        alp_Pc_Dt = self.alp_Pc_Dt
        Fl_Nred = 1.0 / self.Nred

        # Trilinear elements from SIF ELEMENT definitions
        # xa(i,j,m) = x(i,2,m) * x(j,1,m) * k(j,t) where t needs to be determined
        # xb(i,j,m) = x(i,3,m) * x(j,2,m) * k(j,t)
        # xc(i,j,m) = x(i,4,m) * x(j,3,m) * k(j,t)
        # From SIF comment "k(j,EoC)" suggests End of Cycle = last time step
        # Try t = Ntim-1 (last time step, 0-indexed)
        t_eoc = Ntim - 1  # End of Cycle
        xa = jnp.einsum(
            "im,jm,j->ijm", x[:, 1, :], x[:, 0, :], k[:, t_eoc]
        )  # x(i,2,m) * x(j,1,m) * k(j,EoC)
        xb = jnp.einsum(
            "im,jm,j->ijm", x[:, 2, :], x[:, 1, :], k[:, t_eoc]
        )  # x(i,3,m) * x(j,2,m) * k(j,EoC)
        xc = jnp.einsum(
            "im,jm,j->ijm", x[:, 3, :], x[:, 2, :], k[:, t_eoc]
        )  # x(i,4,m) * x(j,3,m) * k(j,EoC)

        # Vectorized constraints

        # Sumi constraints: sum over i of V(i)*x(i,l,m) = 1 for each l,m
        # Shape: (Nage, Ntra)
        sumi = jnp.einsum("i,ilm->lm", V, x) - 1.0
        sumi_flat = sumi.ravel()  # Flatten to 1D

        # Sumlm constraints: sum over l,m of x(i,l,m) = 1 for each i
        # Shape: (Nnod,)
        sumlm = jnp.sum(x, axis=(1, 2)) - 1.0

        # Norm constraints: sum over i of V(i)*k(i,t)*phi(i,t) = 1 for each t
        # Shape: (Ntim,)
        norm = jnp.einsum("i,it,it->t", V, k, phi) - 1.0

        # Kern constraints: keff(t)*phi(i,t) + sum_j G(i,j)*k(j,t)*phi(j,t) = 0
        # Shape: (Nnod, Ntim)
        kern = keff[None, :] * phi + jnp.einsum("ij,jt,jt->it", G, k, phi)
        kern_flat = kern.ravel()  # Flatten to 1D

        # Peak constraints: k(i,t)*phi(i,t) - epsilon/Fl_Nred <= 0
        # Shape: (Nnod, Ntim)
        peak = k * phi - epsilon / Fl_Nred
        peak_flat = peak.ravel()  # Flatten to 1D

        # Burn constraints: k(i,t+1) - k(i,t) + alp_Pc_Dt*k(i,t)*phi(i,t) = 0
        # Shape: (Nnod, Ntim-1)
        burn = k[:, 1:] - k[:, :-1] + alp_Pc_Dt * k[:, :-1] * phi[:, :-1]
        burn_flat = burn.ravel()  # Flatten to 1D

        # Plac constraints: k(i,1) - sum_m x(i,1,m)*Kfresh -
        # sum_{j,m} V(j)*[xa(i,j,m) + xb(i,j,m) + xc(i,j,m)] = 0
        # Note: k(i,1) in SIF is 1-indexed, so k(i,0) in 0-indexed
        # Shape: (Nnod,)
        trilinear_sum = jnp.einsum(
            "j,ijm->i", V, xa + xb + xc
        )  # sum_{j,m} V(j)*[xa + xb + xc]
        plac = k[:, 0] - jnp.sum(x[:, 0, :], axis=1) * self.Kfresh - trilinear_sum

        # Diagonal constraints: x(d1,l,m) = x(d2,l,m) for diagonal pairs
        # From SIF: D1,1=1 D2,1=11, D1,2=7 D2,2=14
        # Python uses 0-indexed, so: (0,10) and (6,13)
        dia1 = x[0, :, :] - x[10, :, :]  # First diagonal pair
        dia2 = x[6, :, :] - x[13, :, :]  # Second diagonal pair
        dia_flat = jnp.concatenate([dia1.ravel(), dia2.ravel()])

        # Kefford constraints: keff(t+1) - keff(t) <= 0
        # Shape: (Ntim-1,)
        kefford = keff[1:] - keff[:-1]

        # Concatenate all equality constraints
        c_eq = jnp.concatenate(
            [sumi_flat, sumlm, norm, kern_flat, burn_flat, plac, dia_flat]
        )

        # Concatenate all inequality constraints
        c_ineq = jnp.concatenate([peak_flat, kefford])

        return c_eq, c_ineq

    @property
    def args(self):
        """No additional arguments."""
        return None

    @property
    def expected_objective_value(self):
        """Expected objective value from SIF file."""
        return None

    @property
    def expected_result(self):
        """Expected result from SIF file."""
        return None


def _build_G_matrix() -> jnp.ndarray:
    """Build the G matrix from SIF values."""
    G = jnp.zeros((14, 14))
    G = G.at[0, 0].set(-0.828)
    G = G.at[1, 0].set(-0.079)
    G = G.at[6, 0].set(-0.014)
    G = G.at[0, 1].set(-0.158)
    G = G.at[1, 1].set(-0.763)
    G = G.at[2, 1].set(-0.079)
    G = G.at[6, 1].set(-0.13)
    G = G.at[7, 1].set(-0.014)
    G = G.at[1, 2].set(-0.079)
    G = G.at[2, 2].set(-0.749)
    G = G.at[3, 2].set(-0.079)
    G = G.at[6, 2].set(-0.028)
    G = G.at[7, 2].set(-0.065)
    G = G.at[8, 2].set(-0.014)
    G = G.at[2, 3].set(-0.079)
    G = G.at[3, 3].set(-0.749)
    G = G.at[4, 3].set(-0.079)
    G = G.at[7, 3].set(-0.014)
    G = G.at[8, 3].set(-0.065)
    G = G.at[9, 3].set(-0.014)
    G = G.at[3, 4].set(-0.079)
    G = G.at[4, 4].set(-0.749)
    G = G.at[5, 4].set(-0.079)
    G = G.at[8, 4].set(-0.014)
    G = G.at[9, 4].set(-0.065)
    G = G.at[4, 5].set(-0.079)
    G = G.at[5, 5].set(-0.749)
    G = G.at[9, 5].set(-0.014)
    G = G.at[0, 6].set(-0.014)
    G = G.at[1, 6].set(-0.065)
    G = G.at[2, 6].set(-0.014)
    G = G.at[6, 6].set(-0.684)
    G = G.at[7, 6].set(-0.065)
    G = G.at[10, 6].set(-0.014)
    G = G.at[1, 7].set(-0.014)
    G = G.at[2, 7].set(-0.065)
    G = G.at[3, 7].set(-0.014)
    G = G.at[6, 7].set(-0.13)
    G = G.at[7, 7].set(-0.698)
    G = G.at[8, 7].set(-0.065)
    G = G.at[10, 7].set(-0.13)
    G = G.at[11, 7].set(-0.014)
    G = G.at[2, 8].set(-0.014)
    G = G.at[3, 8].set(-0.065)
    G = G.at[4, 8].set(-0.014)
    G = G.at[7, 8].set(-0.065)
    G = G.at[8, 8].set(-0.684)
    G = G.at[9, 8].set(-0.065)
    G = G.at[10, 8].set(-0.028)
    G = G.at[11, 8].set(-0.065)
    G = G.at[12, 8].set(-0.014)
    G = G.at[3, 9].set(-0.014)
    G = G.at[4, 9].set(-0.065)
    G = G.at[5, 9].set(-0.014)
    G = G.at[8, 9].set(-0.065)
    G = G.at[9, 9].set(-0.684)
    G = G.at[11, 9].set(-0.014)
    G = G.at[12, 9].set(-0.065)
    G = G.at[6, 10].set(-0.014)
    G = G.at[7, 10].set(-0.065)
    G = G.at[8, 10].set(-0.014)
    G = G.at[10, 10].set(-0.684)
    G = G.at[11, 10].set(-0.065)
    G = G.at[13, 10].set(-0.014)
    G = G.at[7, 11].set(-0.014)
    G = G.at[8, 11].set(-0.065)
    G = G.at[9, 11].set(-0.014)
    G = G.at[10, 11].set(-0.13)
    G = G.at[11, 11].set(-0.698)
    G = G.at[12, 11].set(-0.065)
    G = G.at[13, 11].set(-0.13)
    G = G.at[8, 12].set(-0.014)
    G = G.at[9, 12].set(-0.065)
    G = G.at[11, 12].set(-0.065)
    G = G.at[12, 12].set(-0.684)
    G = G.at[13, 12].set(-0.028)
    G = G.at[10, 13].set(-0.014)
    G = G.at[11, 13].set(-0.065)
    G = G.at[12, 13].set(-0.014)
    G = G.at[13, 13].set(-0.684)
    return G



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class EG2(AbstractUnconstrainedMinimisation):
    """The EG2 function.

    A simple nonlinear problem given as an example in Section 1.2.4 of
    the LANCELOT Manual. The problem is non convex and has several local minima.

    Source:
    A.R. Conn, N. Gould and Ph.L. Toint,
    "LANCELOT, A Fortran Package for Large-Scale Nonlinear Optimization
    (Release A)" Springer Verlag, 1992.

    Note J. Haffner --------------------------------------------------------------------
    Reference: https://doi.org/10.1007/978-3-662-12211-2_1, Chapter 1, page 11
    ------------------------------------------------------------------------------------

    SIF input: N. Gould and Ph. Toint, June 1994.

    Classification: OUR2-AN-1000-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 1000  # Problem specifies N=1000

    def objective(self, y, args):
        # From AMPL: sum {i in 1..N-1} sin(x[1] + x[i]^2 - 1.0) + 0.5*sin(x[N]^2)
        del args
        first = y[0]
        last = y[-1]

        # Sum from i=1 to N-1 of sin(x[1] + x[i]^2 - 1.0)
        # Note: in 0-indexed arrays, this is i=0 to N-2
        f1 = jnp.sum(jnp.sin(first + y[:-1] ** 2 - 1.0))

        # Add 0.5*sin(x[N]^2)
        f2 = 0.5 * jnp.sin(last**2)

        return f1 + f2

    @property
    def y0(self):
        # Initial guess: all zeros (from PyCUTEst)
        return jnp.zeros(self.n)

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        return None

    @property
    def expected_objective_value(self):
        return None



import jax.numpy as jnp

from ..._problem import AbstractBoundedMinimisation
from ..data import load_fbrain_data


# Load FBRAIN data (FBRAIN2LS uses same data as FBRAIN)
A_coeffs, A_lambdas, B_coeffs, B_lambdas, R_values = load_fbrain_data("fbrain")


class FBRAIN2LS(AbstractBoundedMinimisation):
    """FBRAIN2LS - Brain tissue shear stress model with 4 parameters as least squares.

    Match a model of the shear stress in the human brain to data,
    formulated as a nonlinear least-squares problem with 4 parameters.

    Source: an example in
    L.A. Mihai, S. Budday, G.A. Holzapfel, E. Kuhl and A. Goriely.
    "A family of hyperelastic models for human brain tissue",
    Journal of Mechanics and Physics of Solids,
    DOI: 10.1016/j.jmps.2017.05.015 (2017).

    as conveyed by Angela Mihai (U. Cardiff)

    SIF input: Nick Gould, June 2017.

    Classification: SUR2-AN-4-0
    N = 11 (number of data sets)
    M = 200 (number of discretizations)
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0, 1})

    @property
    def y0(self):
        if self.y0_iD == 0:
            return jnp.array([-4.0, -0.1, 4.0, 0.1])
        else:  # y0_iD == 1
            return jnp.array([1.0, 1.0, 1.0, 1.0])

    @property
    def args(self):
        return None

    def objective(self, y, args):
        del args
        alpha1, c01, alpha2, c02 = y

        # Compute betas
        beta1 = 2.0 * alpha1 - 1.0
        beta2 = 2.0 * alpha2 - 1.0

        # Flatten arrays in column-major order to match SIF constraint ordering
        # SIF iterates: DO J (1 to N), DO I (0 to M) -> (I,J) pairs
        # Cast to same dtype as y to avoid type promotion errors
        a_coeffs_flat = A_coeffs.T.ravel().astype(y.dtype)
        a_lambdas_flat = A_lambdas.T.ravel().astype(y.dtype)
        b_coeffs_flat = B_coeffs.T.ravel().astype(y.dtype)
        b_lambdas_flat = B_lambdas.T.ravel().astype(y.dtype)
        r_values_flat = R_values.T.ravel().astype(y.dtype)

        # Compute all 4 element values as per SIF structure:
        # A(I,J) = C01 * AC(I,J) * AL(I,J)^(2*ALPHA1-1)
        # B(I,J) = C01 * BC(I,J) * BL(I,J)^(2*ALPHA1-1)
        # C(I,J) = C02 * AC(I,J) * AL(I,J)^(2*ALPHA2-1)
        # D(I,J) = C02 * BC(I,J) * BL(I,J)^(2*ALPHA2-1)
        a_values = c01 * a_coeffs_flat * (a_lambdas_flat**beta1)
        b_values = c01 * b_coeffs_flat * (b_lambdas_flat**beta1)
        c_values = c02 * a_coeffs_flat * (a_lambdas_flat**beta2)
        d_values = c02 * b_coeffs_flat * (b_lambdas_flat**beta2)

        # Sum all 4 element functions for each (i,j)
        model_values = a_values + b_values + c_values + d_values

        # Compute squared residuals
        residuals = model_values - r_values_flat

        # Return sum of squares (least squares formulation)
        return jnp.sum(residuals**2)

    @property
    def bounds(self):
        # SIF bounds: LO 'DEFAULT' -5.0, UP 'DEFAULT' 5.0 (applied to all 4 variables)
        lower = jnp.array([-5.0, -5.0, -5.0, -5.0])
        upper = jnp.array([5.0, 5.0, 5.0, 5.0])
        return lower, upper

    @property
    def expected_result(self):
        return None

    @property
    def expected_objective_value(self):
        return jnp.array(0.0)
